from __future__ import absolute_import
from Ganga.GPIDev.Schema import Schema, Version, SimpleItem, ComponentItem
from fnmatch import fnmatch
from Ganga.GPIDev.Adapters.IGangaFile import IGangaFile
from Ganga.Utility.logging import getLogger
from Ganga.GPIDev.Base.Proxy import isType, GPIProxyObjectFactory
from Ganga.Utility.Config import getConfig
import re
import copy
import glob
import os
import pickle
import stat
import logging
import Ganga.Utility.Config

logger = getLogger()
regex = re.compile('[*?\[\]]')
badlogger = getLogger('oauth2client.util')
badlogger.setLevel(logging.ERROR)

class GoogleFile(IGangaFile):

    """
    The GoogleFile outputfile type allows for files to be directly uploaded, downloaded, removed and restored from the GoogleDrive service.
    It can be used as part of a job to output data directly to GoogleDrive, or standalone through the Ganga interface.

    example job: j=Job(application=Executable(exe=File('/home/hep/hs4011/Tests/testjob.sh'), args=[]),outputfiles=[GoogleFile('TestJob.txt')])

                 j.submit()

                 ### This job will automatically upload the outputfile 'TestJob.txt' to GoogleDrive.

    example of standalone submission:

                 g=GoogleFile('TestFile.txt')

                 g.localDir = '~/TestDirectory'        ### The file's location must be specified for standalone submission

                 g.put()                               ### The put() method uploads the file to GoogleDrive directly

    The GoogleFile outputfile is also compatible with the Dirac backend, making outputfiles from Dirac-run jobs upload directly to GoogleDrive.
    """

    _schema = Schema(Version(1, 1),
                     {'namePattern': SimpleItem(defvalue="", doc='pattern of the file name'),
                      'localDir': SimpleItem(defvalue="", copyable=1,
                                             doc='local dir where the file is stored, used from get and put methods'),
                      'subfiles': ComponentItem(category='gangafiles', defvalue=[], hidden=1,
                                                sequence=1, copyable=0, doc="collected files from the wildcard namePattern"),
                      'failureReason': SimpleItem(defvalue="", copyable=1,
                                                  doc='reason for the upload failure'),
                      'compressed': SimpleItem(defvalue=False, typelist=[bool], protected=0,
                                               doc='wheather the output file should be compressed before sending somewhere'),
                      'downloadURL': SimpleItem(defvalue="", copyable=1, protected=1,
                                                doc='download URL assigned to the file upon upload to GoogleDrive'),
                      'id': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                       doc='GoogleFile ID assigned to file  on upload to GoogleDrive'),
                      'title': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                                   doc='GoogleFile title of the uploaded file'),
                      'GangaFolderId': SimpleItem(defvalue="", copyable=1, hidden=1, protected=1,
                                                  doc='GoogleDrive Ganga folder  ID')
                      })
    _category = 'gangafiles'
    _name = 'GoogleFile'
    _exportmethods = ["get", "put", "remove", "restore", "deleteCredentials"]

    def __init__(self, namePattern=''):
        super(GoogleFile, self).__init__()
        self.namePattern = namePattern
        self.__initialized = False
        
        self.cred_path = os.path.join(getConfig('Configuration')['gangadir'], 'googlecreddata.pkl')

    def __initializeCred(self):
        while os.path.isfile(self.cred_path) == False:
            from oauth2client.client import OAuth2WebServerFlow

            # Copy your credentials from the APIs Console
#            CLIENT_ID = "54459939297.apps.googleusercontent.com"
#            CLIENT_SECRET = "mAToHx5RpXtwkeYR6nOIe_Yw"
            CLIENT_ID = '776655306197-dirtoquqsm7cpqgepvamofg5t2b5f637.apps.googleusercontent.com'
            CLIENT_SECRET = 'GpdEP-OBZZQLB3k-xxOpzFQG'
            # Check https://developers.google.com/drive/scopes for all
            # available scopes
            OAUTH_SCOPE = 'https://www.googleapis.com/auth/drive.file'

            # Redirect URI for installed apps
            REDIRECT_URI = 'urn:ietf:wg:oauth:2.0:oob'

            # Run through the OAuth flow and retrieve credentials
            credentials = ''
            flow = OAuth2WebServerFlow(
                CLIENT_ID, CLIENT_SECRET, OAUTH_SCOPE, REDIRECT_URI)
            authorize_url = flow.step1_get_authorize_url()
            try:
                import webbrowser
                webbrowser.get('macosx').open(authorize_url, 0, True)
            except:
                try:
                    import webbrowser
                    webbrowser.get(
                        'windows-default').open(authorize_url, 0, True)
                except:
                    try:
                        import webbrowser
                        webbrowser.get('firefox').open(authorize_url, 0, True)
                    except Exception, err:
                        logger.error("Error: %s" % str(err))
                        pass
            logger.info(
                'Go to the following link in your browser: ' + authorize_url)
            code = raw_input('Enter verification code: ').strip()
            try:
                credentials = flow.step2_exchange(code)
            except:
                deny = raw_input(
                    'An incorrect code was entered. Have you denied Ganga access to your GoogleDrive (y/[n])?')
                if deny == '' or deny[0:1].upper() == 'N':
                    pass
                elif deny[0:1].upper() == 'Y':
                    return None

            # Pickle credential data
            if credentials is not '':
                with open(self.cred_path, "wb") as output:
                    pickle.dump(credentials, output)
                
                os.chmod(self.cred_path, stat.S_IWUSR | stat.S_IRUSR)
                logger.info('Your GoogleDrive credentials have been stored in the file %s and are only readable by you. '
                            'The file will give permission to modify files in your GoogleDrive. '
                            'Permission can be revoked by going to "Manage Apps" in your GoogleDrive '
                            'or by deleting the credentials through the deleteCredentials GoogleFile method.' % self.cred_path)

        self.__initialized = True

        self._check_Ganga_folder()

    def __construct__(self, args):
        if (len(args) != 1) or (not isType(args[0], str)):
            super(GoogleFile, self).__construct__(args)
        else:
            self.namePattern = args[0]
        self.cred_path = os.path.join(getConfig('Configuration')['gangadir'], 'googlecreddata.pkl')

    def _attribute_filter__set__(self, n, v):
        if n == 'localDir':
            return os.path.expanduser(os.path.expandvars(v))
        return v

    def setLocation(self):
        """
        Sets the location of output files that were uploaded from the WN
        """
        raise NotImplementedError

    def location(self):
        """
        Return list with the locations of the post processed files (if they were configured to upload the output somewhere)
        """
        raise NotImplementedError

    def _on_attribute__set__(self, obj_type, attrib_name):
        r = copy.deepcopy(self)
        from Ganga.GPIDev.Lib.Job import Job
        if isinstance(obj_type, Job) and attrib_name == 'outputfiles':
            r.localDir = None
            r.failureReason = ''
        return r

    def deleteCredentials(self):
        """
        Deletes the user's GoogleDrive credentials

            example use: GoogleFile().deleteCredentials()
        """
        if self.__initilized == True:
            if os.path.isfile(self.cred_path) == True:
                os.remove(self.cred_path)
                logger.info('GoogleDrive credentials deleted')
                return None
        else:
            logger.info('There are no credentials to delete')

    def get(self):
        """
        Retrieves files uploaded to GoogleDrive through a job or by a standalone GoogleFile
        """
        service = self._setup_service()

        # Checks for wildcards and loops through get procedure for each result,
        # saving file to assigned directory
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                if f.downloadURL:
                    resp, content = service._http.request(f.downloadURL)
                    if resp.status == 200:
                        # print 'Status: %s' % resp
                        logger.info(
                            "File \'%s\' downloaded succesfully" % f.title)
                        dir_path = f.localDir
                        if f.localDir == '':
                            dir_path = self.localDir
                            if self.localDir == '':
                                dir_path = os.getcwd()
                        completeName = os.path.join(dir_path, f.title)
                        with open(completeName, "wb") as gotfile:
                            gotfile.write(content)

                    else:
                        # print 'An error occurred: %s' % resp
                        logger.info(
                            "Download unsuccessful, file \'%s\' may not exist on GoogleDrive" % f.title)
                else:
                    # The file doesn't have any content stored on Drive.
                    logger.info(
                        "No file \'%s\' exists on GoogleDrive" % f.title)
                    return None

        # Non-wildcard get request procedure
        else:
            if self.downloadURL:
                resp, content = service._http.request(self.downloadURL)
                if resp.status == 200:
                    # print 'Status: %s' % resp
                    logger.info("Download successful")
                    dir_path = self.localDir
                    if self.localDir == ('' or None):
                        dir_path = os.getcwd()
                    if self._getParent() is not None:
                        dir_path = self.getJobObject(
                        ).getOutputWorkspace().getPath()
                    completeName = os.path.join(dir_path, self.namePattern)
                    with open(completeName, "wb") as gotfile:
                        gotfile.write(content)
                else:
                    # print 'An error occurred: %s' % resp
                    logger.info(
                        "Download unsuccessful, the file may not exist on GoogleDrive")
                    return None
            else:
                # The file doesn't have any content stored on Drive.
                logger.info("No such file on GoogleDrive")
                return

    def getWNScriptDownloadCommand(self, indent):
        """
        Gets the command used to download already uploaded file
        """
        raise NotImplementedError

    def processWildcardMatches(self):
        raise NotImplementedError

    def __repr__(self):
        """
        Get the representation of the file
        """
        return "GoogleFile(namePattern='%s', downloadURL='%s')" % (self.namePattern, self.downloadURL)

    def put(self):
        """
        Postprocesses (upload) output file to the desired destination from the client
        """
        import hashlib
        from apiclient.http import MediaFileUpload

        service = self._setup_service()

        # Sets the target directory
        dir_path = self.localDir
        if self.localDir == '':
            dir_path = os.getcwd()

        if self._getParent() is not None:
            dir_path = self.getJobObject().getOutputWorkspace().getPath()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for wildfile in glob.glob(os.path.join(dir_path, self.namePattern)):
                FILENAME = wildfile
                filename = os.path.basename(wildfile)

                # Upload procedure
                media_body = MediaFileUpload(
                    FILENAME, mimetype='text/plain', resumable=True)
                body = {
                    'title': '%s' % filename,
                    'description': 'A test document',
                    'mimeType': 'text/plain',
                    'parents': [{
                        "kind": "drive#fileLink",
                        "id": "%s" % self.GangaFolderId
                    }]
                }

                # Metadata file and md5checksum intergrity check
                file = service.files().insert(
                    body=body, media_body=media_body).execute()
                with open(FILENAME, 'rb') as thefile:
                    if file.get('md5Checksum') == hashlib.md5(thefile.read()).hexdigest():
                        logger.info("File \'%s\' uploaded successfully" % filename)
                    else:
                        logger.error(
                            "File \'%s\' uploaded unsuccessfully" % filename)

                # Assign new schema components to each file and append to job
                # subfiles
                g = GoogleFile(filename)
                g.downloadURL = file.get('downloadUrl', '')
                g.id = file.get('id', '')
                g.title = file.get('title', '')
                self.subfiles.append(GPIProxyObjectFactory(g))

        # For non-wildcard upload
        else:
            # Path to the file to upload
            FILENAME = os.path.join(dir_path, self.namePattern)

            # Upload procedure, can edit more of file metadata
            media_body = MediaFileUpload(
                FILENAME, mimetype='text/plain', resumable=True)
            body = {
                'title': '%s' % self.namePattern,
                'description': 'A test document',
                'mimeType': 'text/plain',
                'parents': [{
                    "kind": "drive#fileLink",
                    "id": "%s" % self.GangaFolderId
                }]
            }

            # Metadata storage and md5checksum integrity check
            file = service.files().insert(
                body=body, media_body=media_body).execute()

            with open(FILENAME, 'rb') as thefile:
                if file.get('md5Checksum') == hashlib.md5(thefile.read()).hexdigest():
                    logger.info("File \'%s\' uploaded succesfully" %
                                self.namePattern)
                else:
                    logger.error("Upload Unsuccessful")

            # Assign values to new schema components
            self.downloadURL = file.get('downloadUrl', '')
            self.id = file.get('id', '')
            self.title = file.get('title', '')

            return
        return GPIProxyObjectFactory(self.subfiles[:])

    def remove(self, permanent=False):
        """
        Move a file to the trash or permanently delete the file

            example use: GoogleFile().remove()

            or:          j = Job([...], outputfiles=GoogleFile()) --> j.submit --> j.outputfiles[0].remove()

        Remove multiple files by using

                         for i in j.outputfiles:
                             i.remove()

        The file can also be permanently deleted by using

                         GoogleFile().remove(True)

        However, this will make the file unrestorable
        """
        service = self._setup_service()

        from apiclient import errors

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                if permanent == True:
                    try:
                        service.files().delete(fileId=f.id).execute()
                        f.downloadURL = ''
                        logger.info(
                            'File \'%s\' permanently deleted from GoogleDrive' % f.title)
                    except errors.HttpError as error:
                        # print 'An error occurred: %s' % error
                        logger.info(
                            'File \'%s\' deletion failed, or file already deleted' % f.title)
                else:
                    try:
                        service.files().trash(fileId=f.id).execute()
                        logger.info(
                            'File \'%s\' removed from GoogleDrive' % f.title)
                    except errors.HttpError as error:
                        # print 'An error occurred: %s' % error
                        logger.info(
                            'File \'%s\' removal failed, or file already removed' % f.title)

        # Non-wildcard request
        else:
            if permanent == True:
                try:
                    service.files().delete(fileId=self.id).execute()
                    self.downloadURL = ''
                    logger.info('File permanently deleted from GoogleDrive')
                except errors.HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info(
                        'File deletion failed, or file already deleted')
            else:
                try:
                    service.files().trash(fileId=self.id).execute()
                    logger.info('File removed from GoogleDrive')
                except errors.HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info('File removal failed, or file already removed')
                return None

    def restore(self):
        """
        Restore a file from the trash. This method will not work on permanently deleted files

            example use: GoogleFile().restore()
        """
        from apiclient import errors

        service = self._setup_service()

        # Wildcard procedure
        if regex.search(self.namePattern) is not None:
            for f in self.subfiles:
                try:
                    service.files().untrash(fileId=f.id).execute()
                    logger.info(
                        'File \'%s\' restored to GoogleDrive' % f.title)
                except errors.HttpError as error:
                    # print 'An error occurred: %s' % error
                    logger.info(
                        'File \'%s\' restore failed, or file does not exist on GoogleDrive' % f.title)

        # Non-wildcard request
        else:
            try:
                service.files().untrash(fileId=self.id).execute()
                logger.info('File restored to GoogleDrive')
            except errors.HttpError as error:
                # print 'An error occurred: %s' % error
                logger.info(
                    'File restore failed, or file does not exist on GoogleDrive')
            return None

    def _check_Ganga_folder(self):
        """
        Creates a Ganga folder on GoogleDrive if one is not already present
        """
        from apiclient import errors
        service = self._setup_service()

        page_token = None
        try:
            param = {}
            if page_token:
                param['pageToken'] = page_token
            files = service.files().list(**param).execute()
            items = files['items']
            for i in items:
                if i['title'] == 'Ganga':
                    self.GangaFolderId = i['id']
                    return
            page_token = files.get('nextPageToken')
        except errors.HttpError as error:
            logger.info('Failed to create Ganga folder on GoogleDrive')
            # print 'An error occurred: %s' % error

        body = {
            'title': 'Ganga',
            'description': 'A test folder',
            'mimeType': 'application/vnd.google-apps.folder'
        }
        file = service.files().insert(body=body).execute()
        self.GangaFolderId = file.get('id')

    def _setup_service(self):
        """
        Sets up the GoogleDrive service for other methods
        """
        from apiclient.discovery import build
        import httplib2
        http = httplib2.Http()
        if self.__initialized == False:
            self.__initializeCred()
        with open(self.cred_path, "rb") as nput:
            credentials = pickle.load(nput)
        http = credentials.authorize(http)
        service = build('drive', 'v2', http=http)
        return service

    def getWNInjectedScript(self, outputFiles, indent, patternsToZip, postProcessLocationsFP):
        """
        Returns script that have to be injected in the jobscript for postprocessing on the WN
        """
        logger.info('injecting')

    def _readonly(self):
        return False

    def _list_get__match__(self, to_match):
        if isinstance(to_match, str):
            return fnmatch(self.namePattern, to_match)
        if isinstance(to_match, type):
            # note stripProxy wont work on class types that aren't instances
            return isinstance(self, to_match._impl)
        return to_match == self

Ganga.Utility.Config.config_scope['GoogleFile'] = GoogleFile



################################################################################
# Ganga Project. http://cern.ch/ganga
#
################################################################################

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Adapters.IPostProcessor import PostProcessException
from Ganga.GPIDev.Adapters.IChecker import IChecker, IFileChecker
from Ganga.GPIDev.Base.Proxy import GPIProxyObject
from Ganga.GPIDev.Schema import ComponentItem, FileItem, Schema, SimpleItem, Version
from Ganga.Utility.Config import makeConfig, ConfigError, getConfig
from Ganga.Utility.Plugin import allPlugins
from Ganga.Utility.logging import getLogger, log_user_exception
import commands
import copy
import os
import string
import shutil

# Simon's post_status - communicates to processing DB
import post_status
import urllib2

# later that will taken from Job.Application DVtmp
site = 'wg-bugaboo'
trig = 'SPILL'


logger = getLogger()

class FileCheckeR(IFileChecker):
    """
    Checks if string is in file.
    self.searchStrings are the search strings you would like to check for.
    self.files are the files you would like to check.
    self.failIfFound (default = True) decides whether to fail the job if the string is found. If you set this to false the job will fail if the string *isnt* found.
    self.fileMustExist toggles whether to fail the job if the specified file doesn't exist (default is True).
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['searchStrings'] = SimpleItem(defvalue = [], doc='String to search for')
    _schema.datadict['failIfFound'] = SimpleItem(True, doc='Toggle whether job fails if string is found or not found.')
    _category = 'postprocessor'
    _name = 'FileCheckeR'
    _exportmethods = ['check']


    def check(self,job):
        """
        Check that a string is in a file, takes the job object as input.
        """

        if not len(self.searchStrings):
            raise PostProcessException('No searchStrings specified, FileCheckeR will do nothing!')
        filepaths = self.findFiles(job)
        if not len(filepaths):
            raise PostProcessException('None of the files to check exist, FileCheckeR will do nothing!') 
        for filepath in filepaths:
            for searchString in self.searchStrings:
                grepoutput = commands.getoutput('grep "%s" %s' % (searchString,filepath))
                if len(grepoutput) and self.failIfFound is True:            
                    logger.info('The string %s has been found in file %s, FileCheckeR will fail job(%s)',searchString,filepath,job.fqid)
                    return self.failure
                if not len(grepoutput) and self.failIfFound is False:            
                    logger.info('The string %s has not been found in file %s, FileCheckeR will fail job(%s)',searchString,filepath,job.fqid)
                    return self.failure
        return self.result 


class ND280RDP_Checker(IFileChecker):
    """
    o Checks .log file in a comprehensive way
    o Posts results to processing DB
    o Moves output files to there destinations
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['prfx'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Path prefix to store output files')
    _schema.datadict['path'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Middle path to store output files')
    _schema.datadict['trig'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Trigger type, e.g. SPILL or COSMIC')
    _schema.datadict['site'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Processing site, e.g. wg-bugaboo')
    _category = 'postprocessor'
    _name = 'ND280RDP_Checker'
    _exportmethods = ['check']

    def __init__(self):
        super(ND280RDP_Checker,self).__init__()
        self.filesMustExist = False

        self.InStage = 0
        self.ReturnCode = -1
        self.EventsIn  = 0
        self.EventsOut = 0
        self.Time = 0
        self.STAGE = ''

        self.line = ''

        self.logf = ''
        self.RUN = ''
        self.SUBRUN = ''
        self.TRIGTYPE = None
        
    def find(self,str):
        return self.line.find(str)>=0

    def reset_variables(self):
        self.InStage = 0
        self.ReturnCode = 1
        self.EventsIn  = 0
        self.EventsOut = 0
        self.Time = 0
        self.STAGE = ''

    def send_status(self):
        logger.info('Result for %s %s %s %s is: %s, %s, %s, %s, %s' %  (self.RUN,self.SUBRUN,self.TRIGTYPE,self.STAGE,self.site,self.ReturnCode,self.Time,self.EventsIn,self.EventsOut))

        if not self.path:
            logger.error("No monitoring info sent because MONDIR is not defined")
            return

        rang = self.RUN[:5]+'000_'+self.RUN[:5]+'999'
        mondir = os.path.join(self.path,rang)
        job = { 'run':int(self.RUN), 'subrun':int(self.SUBRUN), 'trigtype':self.TRIGTYPE, 'stage':self.STAGE }
        attributes = {'site':self.site, 'result':self.ReturnCode,
                      'time':int(self.Time), 'read':self.EventsIn, 'written':self.EventsOut }

        try:
            #print mondir, job, attributes
            logger.info("%s %s %s",str(mondir), str(job), str(attributes))
            post_status.record(mondir, job, attributes)
        except urllib2.HTTPError, e:
            #print "Error: The HTTP request failed. (%s)" % (str(e), )
            logger.error("Error: The HTTP request failed. (%s)" % (str(e), ))
            return 4

        return 0

    def check(self,job):

        IsMC = 0
        IsGenie = 0
        IsNeut = 0
        if self.trig == 'SPILL':
            self.TRIGTYPE = 'spill'
        elif self.trig == 'COSMIC':
            self.TRIGTYPE = 'cosmic'
        elif self.trig == 'MCP':
            self.TRIGTYPE = 'spill'
            isMC = 1
        elif self.trig == 'MCCOS':
            self.TRIGTYPE = 'all'
            isMC = 1
        else:
            #print "Unknown type of data: "+self.trig
            #self.move_outs(job,ok=False)
            #return False
            raise PostProcessException("Unknown type of data: "+self.trig)

        if not self.site:
            #print 'Site is not given'
            #self.move_outs(job,ok=False)
            #return False
            raise PostProcessException('Site is not given')
            

        # finds .log file
        self.files = ['*.log']
        filepaths = self.findFiles(job)
        if len(filepaths) != 1:
            #print 'Something wrong with logfile(s) '+filepaths+'. CANNOT CONTINUE'
            logger.error('Something wrong with logfile(s) '+filepaths+'. CANNOT CONTINUE')
            self.move_output(job,ok=False)
            return False
        self.logf = filepaths[0]

        filename = os.path.basename(self.logf)
        chunks = filename.split('_')
        chunks = chunks[3].split('-')
        self.RUN = chunks[0]
        self.SUBRUN = chunks[1]

        if not os.path.exists(self.logf):
            #print "Log file "+filename+" not found. Exit on error"
            logger.error("Log file "+filename+" not found. Exit on error")
            self.move_output(job,ok=False)
            return False

        # Check generator if this is a beam MC
        if IsMC == 1:
            if filename.find('oa_nt_')>=0:
                #print "This is a NEUT MC log file"
                logger.info("This is a NEUT MC log file")
                IsNeut = 1
            elif filename.find('oa_gn_')>=0:
                #print "This is a GENIE MC log file"
                logger.info("This is a GENIE MC log file")
                IsGenie = 1


        #print "Starting to scan file "+filename
        #print "for run %s, subrun %s, type %s" % (self.RUN,self.SUBRUN,self.trig)
        logger.info("Starting to scan file "+filename)
        logger.info("for run %s, subrun %s, type %s" % (self.RUN,self.SUBRUN,self.trig))
        
        inlogf = open(self.logf)
        for self.line in inlogf:
            self.line = self.line.strip('\n')
            if self.find('Midas File') and self.find('has been truncated'):
                #print self.line
                #print "Midas file probably missing"
                logger.error('%s\n%s',self.line,"Midas file probably missing")
                self.ReturnCode = -1
                self.STAGE = "cali"
                self.send_status()
                self.InStage = 0
                break
            
            elif self.find('Starting job for nd280MC'):
                self.InStage = 1
                self.STAGE = "nd280MC"
                #print self.line
                logger.info(self.line)

            elif self.find('Starting job for elecSim'):
                self.InStage = 1
                self.STAGE = "elecSim"
                #print self.line
                logger.info(self.line)

            elif self.find('Starting job for oaCosmicTrigger'):
                self.InStage = 1
                self.STAGE = "COSMICTRIG"
                #print self.line
                logger.info(self.line)

            elif self.find('Starting job for oaCalib'):
                self.InStage = 1
                self.STAGE = "cali"
                #print self.line
                logger.info(self.line)

            elif self.find('Starting job for oaRecon'):
                self.InStage = 1
                self.STAGE = "reco"
                #print self.line
                logger.info(self.line)
 
            elif self.find('Starting job for oaAnalysis'):
                self.InStage = 1
                self.STAGE = "anal"
                #print self.line
                logger.info(self.line)

            elif self.find('Starting job for '):
                #print self.line
                logger.info(self.line)

            elif self.find('Found Command event_select '):
                #print self.line
                logger.info(self.line)
                chunks = self.line.split()
                self.TRIGTYPE = chunks[5]

            elif self.InStage == 1:
                if   self.find('Segmentation fault'):
                    #print self.line
                    logger.error(self.line)
                    if self.line == '"oaCherryPicker-geo_v5mr.bat: line 7"':
                        #print "This is an acceptable error - ignore it"
                        logger.warning("This is an acceptable error - ignore it")
                    else:
                        self.ReturnCode = -2
                        self.InStage = 0
                        self.send_status()
                        break

                elif self.find('Disk quota exceeded'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -3
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find(' ERROR: No database for spillnum'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -8
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find(' No BSD data available'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -8
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('No pass through information could be found'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -10
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('Disabling module '):
                    #print "IsMC = "+str(IsMC)
                    logger.info("IsMC = "+str(IsMC))
                    if IsMC == 1:
                        #print "After testing IsMC = "+str(IsMC)
                        logger.info("After testing IsMC = "+str(IsMC))
                        if self.find('Disabling module GRooTrackerVtx') and IsGenie:
                            #print "Atest "+self.line
                            logger.error("Atest "+self.line)
                            self.ReturnCode = -4
                            self.InStage = 0
                            self.send_status()
                            break
                        elif self.find('Disabling module GRooTrackerVtx') and IsNeut:
                            #print "Btest "+self.line
                            logger.error("Btest "+self.line)
                            self.ReturnCode = -4
                            self.InStage = 0
                            self.send_status()
                            break
                        elif not self.find('RooTracker'):
                            #print "Ctest "+self.line
                            logger.error("Ctest "+self.line)
                            self.ReturnCode = -4
                            self.InStage = 0
                            self.send_status()
                            break

                elif self.find('probably not closed, trying to recover'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -6
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('St9bad_alloc'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -7
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('No luck connecting to GSC MySQL server'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -9
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('EProductionException'):
                    #print self.line
                    logger.error(self.line)
                    self.ReturnCode = -11
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('Total Events Read'):
                    chunks = self.line.split()
                    self.EventsIn = chunks[3]

                elif self.find('Total Events Written'):
                    chunks = self.line.split()
                    self.EventsOut = chunks[3]

                elif self.find('Number of events ='):
                    chunks = self.line.split()
                    if self.STAGE == 'nd280MC':
                        self.EventsOut = chunks[5]

                elif self.find('Total number of events processed in Analysis'):
                    chunks = self.line.split()
                    self.EventsOut = chunks[9]

                elif self.find('Job Completed Successfully'):
                    #print self.line
                    logger.info(self.line)
                    nextline = inlogf.next()
                    if nextline.find('Run time')>=0:
                        chunks = nextline.split()
                        Time = chunks[6]
                        chunks = Time.split('.')
                        self.Time = chunks[0]
                    self.ReturnCode = 1
                    if self.STAGE == 'COSMICTRIG':
                        #print "Not sure what this stage is yet - no call to post_status"
                        logger.warning("Not sure what this stage is yet - no call to post_status")
                    else:
                        self.send_status()
                    self.reset_variables()
                    self.InStage = 0

        if self.InStage == 1:
            #print "The stage "+self.STAGE+" has not completed succesfully, Error is unknown"
            logger.error("The stage "+self.STAGE+" has not completed succesfully, Error is unknown")
            self.ReturnCode = 0
            self.send_status()
            
        inlogf.close()
        #print "\nFinished scanning the log file. Last check return code posted is "+str(self.ReturnCode)
        logger.info("Finished scanning the log file. Last check return code posted is "+str(self.ReturnCode))

        self.move_output(job)
        return self.ReturnCode == 1
    
    def move_output(self,job,ok=True):
        if ok:
            rang = self.RUN[:5]+'000_'+self.RUN[:5]+'999'
            dest = os.path.join(self.prfx,self.path,rang)
            jout = self.trig+'_'+self.RUN+'_'+self.SUBRUN
        else:
            dest = os.path.join(self.prfx,self.path)
            jout = 'stdout'
            
        task = {'*.log':'logf','*_cali_*.root':'cali','*_reco_*.root':'reco','*_anal_*.root':'anal','*catalogue.dat':'cata','stdout':'jobOutput'}
        if self.ReturnCode != 1 or not ok:
            task = {'*.log':'errors','*.root':'errors','*catalogue.dat':'errors','stdout':'errors'}
        for p in task.keys():
            odir = os.path.join(dest,task[p])
            if not os.path.isdir(odir):
                os.makedirs(odir)

            self.files = [p]
            filepaths = self.findFiles(job)
            for f in filepaths:
                shutil.move(f,odir)

                if p == 'stdout':
                    shutil.move(os.path.join(odir,p),os.path.join(odir,jout))



#
# Dictionary with all parameters available for the
# section User.
#
# 08/06/10 @ ubeda
#

from Ganga.GPIDev.Base import GangaObject
from Ganga.GPIDev.Schema import *

class USER(GangaObject):

    _comments = []
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#additional_input_files')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#script_exe')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#script_arguments')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#ui_working_dir')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#thresholdlevel')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#email')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#email')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#return_data__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#outputdir')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#logdir')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#copy_data__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#storage_element')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#user_remote_dir')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#storage_path')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#storage_pool')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#storage_port')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#local_stage_out__')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#publish_data_')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#publish_data_name')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#dbs_url_for_publication')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#publish_zero_event')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#srm_version')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#xml_report')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#usenamespace')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#debug_wrapper')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#deep_debug')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#dontcheckspaceleft')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#check_user_remote_dir')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#tasktype')
    _comments.append('http://belforte.home.cern.ch/belforte/misc/test/crab-v2.7.2.html#ssh_control_persist')

    schemadic = {} 
    schemadic['additional_input_files']  = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[0])
    schemadic['script_exe']              = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[1])
    schemadic['script_arguments']        = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[2])
    schemadic['ui_working_dir']          = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[3])
    schemadic['thresholdLevel']          = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[4])
    schemadic['eMail']                   = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[5])
    schemadic['client']                  = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[6])
    schemadic['return_data']             = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[7])
    schemadic['outputdir']               = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[8])
    schemadic['logdir']                  = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[9])
    schemadic['copy_data']               = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[10])
    schemadic['storage_element']         = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[11])
    schemadic['user_remote_dir']         = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[12])
    schemadic['storage_path']            = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[13])
    schemadic['storage_pool']            = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[14])
    schemadic['storage_port']            = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[15])
    schemadic['local_stage_out']         = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[16])
    schemadic['publish_data']            = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[17])
    schemadic['publish_data_name']       = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[18])
    schemadic['dbs_url_for_publication'] = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[19])
    schemadic['publish_zero_event']      = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[20])
    schemadic['srm_version']             = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[21])
    schemadic['xml_report']              = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[22])
    schemadic['usenamespace']            = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[23])
    schemadic['debug_wrapper']           = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[24])
    schemadic['deep_debug']              = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[25])
    schemadic['dontCheckSpaceLeft']      = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[26])
    schemadic['check_user_remote_dir']   = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[27])
    schemadic['tasktype']                = SimpleItem(defvalue=None, typelist=['type(None)','int'], doc=_comments[28])
    schemadic['ssh_control_persist']     = SimpleItem(defvalue=None, typelist=['type(None)','str'], doc=_comments[29])


    _schema =  Schema(Version(0,0), {})
    _hidden = 1
 



from __future__ import print_function

import Ganga.Utility.logging
from Ganga.GPIDev.Base.Proxy import stripProxy, getName

logger = Ganga.Utility.logging.getLogger()


def _initconfigFeed():
    """Initialize Feedback configuration."""
    try:
        from Ganga.Utility.Config import getConfig, ConfigError
        config = getConfig("Feedback")

        def deny_modification(name, x):
            raise ConfigError(
                'Cannot modify [Feedback] settings (attempted %s=%s)' % (name, x))
        config.attachUserHandler(deny_modification, None)
    except ImportError as err:
        # on worker node so Config is not needed since it is copied to Feedback
        # constructor
        logger.debug("Import Error: %s" % err)
        pass
_initconfigFeed()


def report(job=None):
    """ Upload error reports (snapshot of configuration,job parameters, input/output files, command history etc.). Job argument is optional. """
    import mimetypes
    import urllib
    import urllib2
    import httplib
    import string
    import random
    import sys
    import os
    import platform

    import Ganga.GPIDev.Lib.Config.config as config
    from Ganga.GPIDev.Base.VPrinter import full_print

    import Ganga

    # global variables that will print sumamry report to the user along with
    # the download link
    global JOB_REPORT, GANGA_VERSION, BACKEND_NAME, APPLICATION_NAME, PYTHON_PATH
    JOB_REPORT = False
    GANGA_VERSION = ''
    BACKEND_NAME = ''
    APPLICATION_NAME = ''
    PYTHON_PATH = ''

    def random_string(length):
        return ''.join([random.choice(string.letters) for ii in range(length + 1)])

    def encode_multipart_formdata(files):
        boundary = random_string(30)
        retnl = '\r\n'
        lines = []

        def get_content_type(filename):
            return mimetypes.guess_type(filename)[0] or 'application/octet-stream'

        fields = {'title': 'Ganga Error Report'}

        for (key, value) in fields.iteritems():
            lines.append('--' + boundary)
            lines.append('Content-Disposition: form-data; name="%s"' % key)
            lines.append('')
            lines.append(value)
        for field_name, file in files.iteritems():
            lines.append('--' + boundary)
            lines.append(
                'Content-Disposition: form-data; name="file"; filename="%s"' % (file))
            lines.append('Content-Type: %s' % get_content_type(file))
            lines.append('')
            lines.append(open(file, 'rb').read())
        lines.append('--' + boundary + '--')
        lines.append('')
        body = retnl.join(lines)

        headers = {'content-type': 'multipart/form-data; boundary=%s' %
                   boundary, 'content-length': str(len(body))}

        return body, headers

    def make_upload_file(server):

        def upload_file(path):

            # print 'Uploading %r to %r' % (path, server)

            data = {'MAX_FILE_SIZE': '3145728',
                    'sub': '',
                    'mode': 'regist'}
            files = {'file': path}

            send_post(server, files)

        return upload_file

    def send_post(url, files):
        logger.debug("Sending Post to %s ,  containing %s" % (url, files))

        encoded_data = encode_multipart_formdata(files)

        data = urllib.urlencode(encoded_data[1])
        req = urllib2.Request(url, data=data)
        if req.has_data():
            logger.debug("urllib2: Success!")
        else:
            logger.debug("urllib2: Fail!!!")

        connection = httplib.HTTPConnection(req.get_host())
        # connection.set_debuglevel(1)
        logger.debug("Requesting: 'POST', %s, %s " % (url, encoded_data[1]))
#                connection.request( method='POST', url=req.get_selector(), body=encoded_data[0], headers=encoded_data[1] )
        connection.request(
            method='POST', url=url, body=encoded_data[0], headers=encoded_data[1])
        response = connection.getresponse()

        logger.debug("httplib POST request response was: %s , because: %s" % (
            response.status, response.reason))

        responseResult = response.read()

        #logger.debug("Responce.read(): --%s--" % responseResult )

        responseResult = responseResult[
            responseResult.find("<span id=\"download_path\""):]
        startIndex = responseResult.find("path:") + 5
        endIndex = responseResult.find("</span>")

        logger.debug("Responce.read(): --%s--" %
                     responseResult[startIndex:endIndex])

        logger.info(
            'Your error report was uploaded to ganga developers with the following URL. ')
        logger.info(
            'You may include this URL and the following summary information in your bug report or in the support email to the developers.')
        logger.info('')
        logger.info('***' + str(responseResult[startIndex:endIndex]) + '***')
        logger.info('')
        global GANGA_VERSION, JOB_REPORT, APPLICATION_NAME, BACKEND_NAME, PYTHON_PATH
        logger.info('Ganga Version : ' + GANGA_VERSION)
        logger.info('Python Version : ' + "%s.%s.%s" %
                    (sys.version_info[0], sys.version_info[1], sys.version_info[2]))
        logger.info('Operation System Version : ' + platform.platform())

        if JOB_REPORT:
            logger.info('Application Name : ' + APPLICATION_NAME)
            logger.info('Backend Name : ' + BACKEND_NAME)

        logger.info('Python Path : ' + PYTHON_PATH)
        logger.info('')

        JOB_REPORT = False
        GANGA_VERSION = ''
        BACKEND_NAME = ''
        APPLICATION_NAME = ''
        PYTHON_PATH = ''

    def run_upload(server, path):

        upload_file = make_upload_file(server)
        upload_file(path)

    def report_inner(job=None, isJob=False, isTask=False):

        userInfoDirName = "userreport"
        tempDirName = "reportsRepository"
        # job relevant info
        jobSummaryFileName = "jobsummary.txt"
        jobFullPrintFileName = "jobfullprint.txt"
        repositoryPath = "repository/$usr/LocalXML/6.0/jobs/$thousandsNumxxx"
        # task relevant info
        taskSummaryFileName = "tasksummary.txt"
        taskFullPrintFileName = "taskfullprint.txt"
        tasksRepositoryPath = "repository/$usr/LocalXML/6.0/tasks/$thousandsNumxxx"
        # user's info
        environFileName = "environ.txt"
        userConfigFileName = "userconfig.txt"
        defaultConfigFileName = "gangarc.txt"
        ipythonHistoryFileName = "ipythonhistory.txt"
        gangaLogFileName = "gangalog.txt"
        jobsListFileName = "jobslist.txt"
        tasksListFileName = "taskslist.txt"
        thread_trace_file_name = 'thread_trace.html'
        from Ganga.Utility import Config
        uploadFileServer = Config.getConfig('Feedback')['uploadServer']
        #uploadFileServer= "http://gangamon.cern.ch/django/errorreports/"
        #uploadFileServer= "http://ganga-ai-02.cern.ch/django/errorreports/"
        #uploadFileServer= "http://127.0.0.1:8000/errorreports"

        def printDictionary(dictionary, file=sys.stdout):
            for k, v in dictionary.iteritems():
                print('%s: %s' % (k, v), file=file)

                if k == 'PYTHONPATH':
                    global PYTHON_PATH
                    PYTHON_PATH = v

        def extractFileObjects(fileName, targetDirectoryName):
            try:
                fileToRead = open(fileName, 'r')
                try:
                    fileText = fileToRead.read()
                    import re
                    pattern = "File\(name=\'(.+?)\'"
                    matches = re.findall(pattern, fileText)

                    for fileName in matches:
                        fileName = os.path.expanduser(fileName)
                        targetFileName = os.path.join(
                            targetDirectoryName, os.path.basename(fileName))
                        shutil.copyfile(fileName, targetFileName)

                finally:
                    fileToRead.close()
            # except IOError, OSError:
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

        def writeErrorLog(errorMessage):
            try:
                fileToWrite = open(errorLogPath, 'a')
                try:
                    fileToWrite.write(errorMessage)
                    fileToWrite.write("\n")
                except Exception as err:
                    logger.debug("Err: %s" % err)
                    raise
                finally:
                    fileToWrite.close()
            except Exception as err2:
                logger.debug("Err: %s" % err2)
                pass

        def writeStringToFile(fileName, stringToWrite):

            try:
                # uncomment this to try the error logger
                #fileName = '~/' + fileName
                fileToWrite = open(fileName, 'w')
                try:
                    fileToWrite.write(stringToWrite)
                except Exception as err:
                    logger.debug("Err: %s" % err)
                    raise err
                finally:
                    fileToWrite.close()
            # except IOError:
            except Exception as err:
                logger.debug("Err2: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

        def renameDataFiles(directory):

            for fileName in os.listdir(directory):
                fullFileName = os.path.join(directory, fileName)
                if os.path.isfile(fullFileName):
                    if fileName == 'data':
                        os.rename(fullFileName, fullFileName + '.txt')
                else:
                    renameDataFiles(fullFileName)

        import shutil
        import tarfile
        import tempfile
        import os

        userHomeDir = os.getenv("HOME")
        tempDir = tempfile.mkdtemp()

        errorLogPath = os.path.join(tempDir, 'reportErrorLog.txt')

        fullPathTempDir = os.path.join(tempDir, tempDirName)
        fullLogDirName = ''
        # create temp dir and specific dir for the job/user

        try:
            if not os.path.exists(fullPathTempDir):
                os.mkdir(fullPathTempDir)

            import datetime
            now = datetime.datetime.now()
            userInfoDirName = userInfoDirName + \
                now.strftime("%Y-%m-%d-%H:%M:%S")
            fullLogDirName = os.path.join(fullPathTempDir, userInfoDirName)

            # if report directory exists -> delete it's content(we would like
            # last version of the report)
            if os.path.exists(fullLogDirName):
                shutil.rmtree(fullLogDirName)

            os.mkdir(fullLogDirName)
        # except OSError:
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # import os.environ in a file
        fullEnvironFileName = os.path.join(fullLogDirName, environFileName)

        try:
            inputFile = open(fullEnvironFileName, 'w')
            try:
                printDictionary(os.environ, file=inputFile)

                print('OS VERSION : ' + platform.platform(), file=inputFile)

            finally:
                inputFile.close()
        # except IOError
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # import user config in a file
        userConfigFullFileName = os.path.join(
            fullLogDirName, userConfigFileName)

        try:
            inputFile = open(userConfigFullFileName, 'w')
            try:

                print("#GANGA_VERSION = %s" %
                      config.System.GANGA_VERSION, file=inputFile)

                global GANGA_VERSION
                GANGA_VERSION = config.System.GANGA_VERSION

                # this gets the default values
                # Ganga.GPIDev.Lib.Config.Config.print_config_file()

                # this should get the changed values
                for c in config:
                    print(config[c], file=inputFile)

            finally:
                inputFile.close()
        # except IOError does not catch the exception ???
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # write gangarc - default configuration
        defaultConfigFullFileName = os.path.join(
            fullLogDirName, defaultConfigFileName)

        try:
            outputFile = open(os.path.join(userHomeDir, '.gangarc'), 'r')

            try:
                writeStringToFile(defaultConfigFullFileName, outputFile.read())
            finally:
                outputFile.close()

        # except IOError does not catch the exception ???
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # import ipython history in a file
        try:
            ipythonFile = open(
                os.path.join(os.environ['IPYTHONDIR'], 'history'), 'r')

            try:
                lastIPythonCommands = ipythonFile.readlines()[-20:]
                writeStringToFile(os.path.join(
                    fullLogDirName, ipythonHistoryFileName), '\n'.join(lastIPythonCommands))
                #writeStringToFile(os.path.join(fullLogDirName, ipythonHistoryFileName), ipythonFile.read())
            finally:
                ipythonFile.close()
        # except IOError does not catch the exception ???
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # import gangalog in a file
        userLogFileLocation = config["Logging"]._logfile
        userLogFileLocation = os.path.expanduser(userLogFileLocation)

        try:
            gangaLogFile = open(userLogFileLocation, 'r')
            try:
                writeStringToFile(
                    os.path.join(fullLogDirName, gangaLogFileName), gangaLogFile.read())
            finally:
                gangaLogFile.close()
        # except IOError:
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # import the result of jobs command in the report
        jobsListFullFileName = os.path.join(fullLogDirName, jobsListFileName)

        try:
            outputFile = open(jobsListFullFileName, 'w')
            try:

                from Ganga.Core.GangaRegistry import getRegistryProxy
                print(getRegistryProxy('jobs'), file=outputFile)

            finally:
                outputFile.close()

        # except IOError does not catch the exception ???
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # import the result of tasks command in the report
        tasksListFullFileName = os.path.join(fullLogDirName, tasksListFileName)

        try:
            outputFile = open(tasksListFullFileName, 'w')
            try:

                from Ganga.Core.GangaRegistry import getRegistryProxy
                print(getRegistryProxy('tasks'), file=outputFile)

            finally:
                outputFile.close()

        # except IOError does not catch the exception ???
        except Exception as err:
            logger.debug("Err: %s" % err)
            writeErrorLog(str(sys.exc_info()[1]))

        # save it here because we will change fullLogDirName, but we want this
        # to be the archive and to be deleted
        folderToArchive = fullLogDirName

        # import job relevant info
        if (job is not None and isJob):

            global JOB_REPORT, APPLICATION_NAME, BACKEND_NAME

            JOB_REPORT = True
            APPLICATION_NAME = getName(job.application)
            BACKEND_NAME = getName(job.backend)

            # create job folder
            jobFolder = 'job_%s' % job.fqid
            fullLogDirName = os.path.join(fullLogDirName, jobFolder)
            os.mkdir(fullLogDirName)

            # import job summary in a file
            fullJobSummaryFileName = os.path.join(
                fullLogDirName, jobSummaryFileName)
            writeStringToFile(fullJobSummaryFileName, job)

            # import job full print in a file
            fullJobPrintFileName = os.path.join(
                fullLogDirName, jobFullPrintFileName)

            try:
                inputFile = open(fullJobPrintFileName, 'w')
                try:
                    full_print(job, inputFile)
                finally:
                    inputFile.close()
            # except IOError, OSError:
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

            # extract file objects
            try:
                fileObjectsPath = os.path.join(fullLogDirName, 'fileobjects')
                os.mkdir(fileObjectsPath)
                extractFileObjects(fullJobSummaryFileName, fileObjectsPath)
            # except OSError:
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

            # copy dir of the job ->input/output and subjobs
            try:
                parentDir, currentDir = os.path.split(job.inputdir[:-1])
                workspaceDir = os.path.join(fullLogDirName, 'workspace')
                shutil.copytree(parentDir, workspaceDir)
            # except IOError, OSError
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

            # copy shared area of the job
            try:

                if hasattr(job.application, 'is_prepared'):
                    if job.application.is_prepared is not None and job.application.is_prepared is not True:
                        import os
                        from Ganga.Utility.Config import getConfig
                        from Ganga.Utility.files import expandfilename
                        shared_path = os.path.join(expandfilename(getConfig(
                            'Configuration')['gangadir']), 'shared', getConfig('Configuration')['user'])
                        shareddir = os.path.join(
                            shared_path, job.application.is_prepared.name)
                        if os.path.isdir(shareddir):

                            sharedAreaDir = os.path.join(
                                fullLogDirName, 'sharedarea')
                            shutil.copytree(shareddir, sharedAreaDir)
            # except IOError, OSError
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

            # copy repository job file
            try:
                indexFileName = str(job.id) + '.index'

                repositoryPath = repositoryPath.replace(
                    '$usr', os.getenv("USER"))

                # check if the job is subjob -> different way of forming the
                # path to the repository
                is_subjob = job.fqid.find('.') > -1

                if is_subjob:

                    jobid, subjobid = job.fqid.split(
                        '.')[0], job.fqid.split('.')[1]
                    repositoryPath = repositoryPath.replace(
                        '$thousandsNum', str(int(jobid) / 1000))
                    repositoryPath = os.path.join(repositoryPath, jobid)

                else:
                    repositoryPath = repositoryPath.replace(
                        '$thousandsNum', str(job.id / 1000))

                repositoryFullPath = os.path.join(
                    config.Configuration.gangadir, repositoryPath)
                indexFileSourcePath = os.path.join(
                    repositoryFullPath, indexFileName)
                repositoryFullPath = os.path.join(
                    repositoryFullPath, str(job.id))

                repositoryTargetPath = os.path.join(
                    fullLogDirName, 'repository', str(job.id))

                os.mkdir(os.path.join(fullLogDirName, 'repository'))

                shutil.copytree(repositoryFullPath, repositoryTargetPath)
                # data files are copied but can not be opened -> add .txt to
                # their file names
                renameDataFiles(repositoryTargetPath)

                if not is_subjob:
                    # copy .index file
                    indexFileTargetPath = os.path.join(
                        fullLogDirName, 'repository', indexFileName)
                    shutil.copyfile(indexFileSourcePath, indexFileTargetPath)

            # except OSError, IOError:
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

        # import task relevant info
        if (job is not None and isTask):
            # job is actually a task object
            task = job
            # create task folder
            taskFolder = 'task_%s' % task.id
            fullLogDirName = os.path.join(fullLogDirName, taskFolder)
            os.mkdir(fullLogDirName)

            # import task summary in a file
            fullTaskSummaryFileName = os.path.join(
                fullLogDirName, taskSummaryFileName)
            writeStringToFile(fullTaskSummaryFileName, str(task))

            # import task full print in a file
            fullTaskPrintFileName = os.path.join(
                fullLogDirName, taskFullPrintFileName)

            try:
                inputFile = open(fullTaskPrintFileName, 'w')
                try:
                    full_print(task, inputFile)
                except Exception as err:
                    logger.debug("Err: %s" % err)
                    raise err
                finally:
                    inputFile.close()
            # except IOError, OSError:
            except Exception as err:
                logger.debug("Err2: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

            # copy shared area of the task
            try:
                if len(task.transforms) > 0:
                    if hasattr(task.transforms[0], 'application') and hasattr(task.transforms[0].application, 'is_prepared'):
                        if task.transforms[0].application.is_prepared is not None and task.transforms[0].application.is_prepared is not True:
                            import os
                            from Ganga.Utility.Config import getConfig
                            from Ganga.Utility.files import expandfilename
                            shared_path = os.path.join(expandfilename(getConfig(
                                'Configuration')['gangadir']), 'shared', getConfig('Configuration')['user'])
                            shareddir = os.path.join(
                                shared_path, task.transforms[0].application.is_prepared.name)
                            if os.path.isdir(shareddir):

                                sharedAreaDir = os.path.join(
                                    fullLogDirName, 'sharedarea')
                                shutil.copytree(shareddir, sharedAreaDir)
            # except IOError, OSError
            except Exception as err:
                logger.debug("Err: %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

            # copy repository task file
            try:
                indexFileName = str(task.id) + '.index'

                tasksRepositoryPath = tasksRepositoryPath.replace(
                    '$usr', os.getenv("USER"))
                tasksRepositoryPath = tasksRepositoryPath.replace(
                    '$thousandsNum', str(task.id / 1000))

                repositoryFullPath = os.path.join(
                    config.Configuration.gangadir, tasksRepositoryPath)
                indexFileSourcePath = os.path.join(
                    repositoryFullPath, indexFileName)
                repositoryFullPath = os.path.join(
                    repositoryFullPath, str(task.id))

                repositoryTargetPath = os.path.join(
                    fullLogDirName, 'repository', str(task.id))

                os.mkdir(os.path.join(fullLogDirName, 'repository'))

                shutil.copytree(repositoryFullPath, repositoryTargetPath)
                # data files are copied but can not be opened -> add .txt to
                # their file names
                renameDataFiles(repositoryTargetPath)

                # copy .index file
                indexFileTargetPath = os.path.join(
                    fullLogDirName, 'repository', indexFileName)
                shutil.copyfile(indexFileSourcePath, indexFileTargetPath)

            # except OSError, IOError:
            except Exception as err:
                logger.debug("Err %s" % err)
                writeErrorLog(str(sys.exc_info()[1]))

        # Copy thread stack trace file
        try:
            thread_trace_source_path = os.path.join(getConfig('Configuration')['gangadir'], thread_trace_file_name)
            thread_trace_target_path = os.path.join(fullLogDirName, thread_trace_file_name)
            shutil.copyfile(thread_trace_source_path, thread_trace_target_path)
        except (OSError, IOError) as err:
            logger.debug('Err %s', err)
            writeErrorLog(str(sys.exc_info()[1]))

        resultArchive = '%s.tar.gz' % folderToArchive

        try:
            resultFile = tarfile.TarFile.open(resultArchive, 'w:gz')
            try:
                resultFile.add(
                    folderToArchive, arcname=os.path.basename(folderToArchive))
                # put the error log in the archive
                if(os.path.exists(errorLogPath)):
                    resultFile.add(
                        errorLogPath, arcname=os.path.basename(errorLogPath))
            except Exception as err:
                logger.debug("Err: %s" % err)
                raise
            finally:
                resultFile.close()
        except Exception as err:
            logger.debug("Err2: %s" % err)
            raise  # pass

        # remove temp dir
        if(os.path.exists(folderToArchive)):
            shutil.rmtree(folderToArchive)

        # print the error if there is something
        if os.path.exists(errorLogPath):
            logger.error('')
            logger.error('An error occured while collecting report information : ' + open(errorLogPath, 'r').read())
            logger.error('')

        # delete the errorfile from user's pc
        if(os.path.exists(errorLogPath)):
            os.remove(errorLogPath)

        # return the path to the archive and the path to the upload server
        return (resultArchive, uploadFileServer, tempDir)

    def removeTempFiles(tempDir):
        import shutil

        # remove temp dir
        if os.path.exists(tempDir):
            shutil.rmtree(tempDir)

        # remove temp files from django upload-> if the file is bigger than 2.5
        # mb django internally stores it in tmp file during the upload
        userTempDir = '/tmp/'

        for fileName in os.listdir(userTempDir):
            if fileName.find('.upload') > -1:
                os.remove(os.path.join(userTempDir, fileName))

    tempDir = ''

    # call the report function
    try:
        isJob = isTask = False

        # make typecheck of the param passed
        if job is not None:
            from Ganga.GPIDev.Lib.Job.Job import Job
            from Ganga.GPIDev.Base.Proxy import stripProxy
            isJob = isinstance(stripProxy(job), Job)
            if hasattr(stripProxy(job), '_category') and (stripProxy(job)._category == 'tasks'):
                isTask = True

            if not (isJob or isTask):
                logger.error("report() function argument should be reference to a job or task object")
                return

        resultArchive, uploadFileServer, tempDir = report_inner(
            job, isJob, isTask)

        report_bytes = os.path.getsize(resultArchive)

        if report_bytes > 1024 * 1024 * 100:  # if bigger than 100MB
            logger.error(
                'The report is bigger than 100MB and can not be uploaded')
        else:
            run_upload(server=uploadFileServer, path=resultArchive)

    except Exception as err:
        logger.debug("Err: %s" % err)
        removeTempFiles(tempDir)
        raise  # pass
        # raise

    # removeTempFiles(tempDir)



import os, system, time, datetime, subprocess
from Ganga.Utility.Config import getConfig

config = getConfig('Configuration')
gangadir = config['gangadir']
config = getConfig('System')
gangapath = config['GANGA_PYTHONPATH']

testdir = os.path.join(gangadir, "TestRobot")
os.chdir(testdir)
f = open('heartbeat.txt')
Data = f.readline()
EqualsPoint = -1
for i in range(len(Data)):
    if (Data[i] == '-'):
        EqualsPoint = i
        
if (EqualsPoint != -1):
    LogTime = Data[:(EqualsPoint-1)].strip()
    ProcessID = int(Data[(EqualsPoint+1):].strip())
    LogTime = datetime.datetime(*(time.strptime(LogTime,"%H:%M:%S %j %y"))[0:5])
    MaxTime = datetime.datetime().now() - datetime.timedelta(minutes=30)
if (LogTime < Maxtime):
    os.kill(ProcessID,"")
    cmd = "ganga --config="+gangapath+"/GangaTest/Lib/TestRobot/TESTROBOT.ini robot run"
    subprocess.popen(cmd)
f.close()



import time
import os.path
from Ganga.Utility.Config import getConfig
config = getConfig('Configuration')
#config.addOption('UsageMonitoringURL', "http://gangamon.cern.ch:8888/apmon/ganga.conf",
#                 'MonALISA configuration file used to setup the destination of usage messages')


monitor = None

def _setupMonitor( _URL, msg, msg2 ):
    import ApMon.apmon
    # the ApMon constructor may start background threads to refresh the configuration from URL
    # NOTE: the configuration (including defaultLogLevel) is overriden from
    # the config file specified in URL
    global monitor
    monitor = ApMon.apmon.ApMon(_URL, defaultLogLevel = ApMon.apmon.Logger.FATAL)
    monitor.sendParameters('GangaUsage', msg, msg2)
    # stop any background threads started by the ApMon constructor
    monitor.free()

def ganga_started(session_type, **extended_attributes):
    host = getConfig('System')['GANGA_HOSTNAME']
    version = getConfig('System')['GANGA_VERSION']
    user = getConfig('Configuration')['user']
    runtime_packages = ':'.join(
        map(os.path.basename, filter(lambda x: x, config['RUNTIME_PATH'].split(':'))))
    start = long(time.time() * 1000)

    usage_message = {'user': user, 'host': host, 'start': start, 'session':
                     session_type, 'runtime_packages': runtime_packages, 'version': version}

    usage_message.update(extended_attributes)

    if config['UsageMonitoringMSG']:
        from Ganga.GPIDev.MonitoringServices import MSGUtil
        msg_config = getConfig('MSGMS')
        p = MSGUtil.createPublisher(
            msg_config['server'],
            msg_config['port'],
            msg_config['username'],
            msg_config['password'])
        # start publisher thread and enqueue usage message for sending
        p.start()
        p.send(msg_config['usage_message_destination'], repr(usage_message), {'persistent': 'true'})
        # ask publisher thread to stop. it will send queued message anyway.
        p.stop()
        p._finalize(10.)
        if hasattr(p, 'unregister'):
            p.unregister()
        del p

def ganga_job_submitted(application_name, backend_name, plain_job, master_job, sub_jobs):
    host = getConfig('System')['GANGA_HOSTNAME']
    user = getConfig('Configuration')['user']
    runtime_packages = ':'.join(map(os.path.basename, filter(lambda x: x, config['RUNTIME_PATH'].split(':'))))
    start = long(time.time() * 1000)

    job_submitted_message = {'application': application_name, 'backend': backend_name, 'user': user, 'host': host, 'start':
                             start, 'plain_job': plain_job, 'master_job': master_job, 'sub_jobs': sub_jobs, 'runtime_packages': runtime_packages}

    if config['UsageMonitoringMSG']:
        from Ganga.GPIDev.MonitoringServices import MSGUtil
        msg_config = getConfig('MSGMS')
        p = MSGUtil.createPublisher(
            msg_config['server'],
            msg_config['port'],
            msg_config['username'],
            msg_config['password'])

        # start publisher thread and enqueue usage message for sending
        p.start()
        p.send(msg_config['job_submission_message_destination'], repr(job_submitted_message), {'persistent': 'true'})
        # p.send('/queue/test.ganga.jobsubmission',repr(job_submitted_message),{'persistent':'true'})
        # ask publisher thread to stop. it will send queued message anyway.
        p.stop()
        p._finalize(10.)
        if hasattr(p, 'unregister'):
            p.unregister()
        del p





################################################################################
# GangaND280 Project.
# Anthony Hillairet
# Created 17/01/2014
################################################################################
"""@package ND280Dataset
Ganga module to control the input files.
It relies on the basic ND280Dataset class which is used through inheritance in other more specific classes.
"""


import os, re, fnmatch
import commands

from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Schema import *

from Ganga.Utility.logging import getLogger

logger = getLogger()

class ND280Dataset(Dataset):
    """
    Base class for ND280 Datasets.
        dataset = ND280Dataset()

    You can define the list of files by hand directly if you want:
        dataset.names = ['/path/to/the/files/Input1.root','/path/to/the/files/Input2.root']
    """
    
    _schema = Schema(Version(1,0), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path'),
    })
    
    _category = 'datasets'
    _name = 'ND280Dataset'
    _hidden = 1

    def __init__(self):
        super(ND280Dataset, self).__init__()
        
    def get_raw_from_list(self,prfx,list_file):
       """Get the dataset of raw files as listed in a text file as run/subrun combinations."""

       logger.info('Reading list file %s ...',list_file)

       if not os.path.isdir(prfx):
           logger.error('Directory $s does not exist',prfx)
           return

       if not os.path.exists(list_file):
           logger.error('File %s does not exist',list_file)
           return

       f = open( list_file )
       for ln in f.readlines():

           chunks = ln.split()
           run = "%08d" % int(chunks[0])
           sub = "%04d" % int(chunks[1])
           rang = run[:5]+'000_'+run[:5]+'999'
           file = 'nd280_'+run+'_'+sub+'.daq.mid.gz'
           self.get_dataset( os.path.join(prfx,rang), file)

    def get_dataset_from_list(self,list_file):
       """Get the dataset files as listed in a text file."""

       logger.info('Reading list file %s ...',list_file)

       if not os.path.exists(list_file):
           logger.error('File %s does not exist',list_file)
           return

       f = open( list_file )
       for ln in f.readlines():

           # split the directory from the file and call get_dataset
           if os.path.isdir(ln.strip()):
               self.get_dataset( ln.strip() )
           else:
               self.get_dataset( os.path.dirname(ln.strip()), os.path.basename( ln.strip() ) )
        
    def set_dataset_into_list(self,list_file):
       """Write the dataset files as a list in a text file."""

       logger.info('Writing list file %s ...',list_file)

       if not len(self.names):
         logger.error('No input files found. The dataset is empty.')
         return 0

       try:
         fileList = open( list_file, 'w')
       except IOError:
         logger.error('File %s cannot be created',list_file)
         return 0

       for inFile in self.names:
         fileList.write(inFile+'\n')

       fileList.close()
       return 1

    def get_dataset(self,directory,filter=None):
        """Get the list of files in the dataset's directory.
        This is not implemented in ND280Dataset but is defined in each class inheriting from ND280Dataset."""
      
        raise NotImplementedError


    def get_dataset_filenames(self):
        """Simply returns a python list containing all the filenames in this ND280Dataset.
        """
        return self.names

    def set_dataset_filenames(self,list_file):
       """Copy the list of files given in input as the list of files in the ND280Dataset.
       NOTE: This will not append the input list to the existing file list but instead replace it."""

       logger.info('Writing list file %s ...',list_file)

       self.names=list_file

    def get_filenames(app):
        """Retrieve the file names starting from an application object"""
      
        job=app._getRoot()
        if not job:
            logger.warning('Application object is not associated to a job.')
            return []
         
        # Jobs without inputdata are allowed
        if not job.inputdata: return []
      
        import re
        classnamematch = re.match(r"ND280.*Dataset", job.inputdata._name)
        if not classnamematch:
            logger.warning('Dataset is not a class inheriting from ND280Dataset.')
            return []

        return job.inputdata.names
         
    get_filenames=staticmethod(get_filenames)


class ND280LocalDataset(ND280Dataset):

    """ND280LocalDataset manages files located in a local directory."""
    
    _schema = Schema(Version(1,1), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path'),
    })
    
    _category = 'datasets'
    _name = 'ND280LocalDataset'

    _exportmethods = ['get_dataset', 'get_dataset_filenames', 'get_dataset_from_list', 'get_raw_from_list', 'set_dataset_into_list', 'set_dataset_filenames' ]

    _GUIPrefs = [ { 'attribute' : 'names',  'attribute' :  'String_List' } ]
   
    def __init__(self):
        super(ND280LocalDataset, self).__init__()


    def get_dataset(self,directory,filter=None):
       """Get the list of files in the local directory given.
       You can provide a wildcard filter such as 'oa_*.root'"""
      
       logger.info('Reading %s ...',directory)


       if not os.path.isdir(directory):
           logger.error('Path %s is no directory',directory)
           return

       directory = os.path.abspath(directory)
       if filter:
           new_names = [ os.path.join(directory,name) for name in fnmatch.filter(os.listdir(directory),filter) ]
       else:
           new_names = [ os.path.join(directory,name) for name in os.listdir(directory) ]

       self.names.extend( new_names )

       self._setDirty()


class ND280DCacheDataset(ND280Dataset):

    """ND280 local datasets manages files located in a directory on a DCache server.
    By default, the configured server is TRIUMF.
    And currently the only configured server is TRIUMF but later you will be able to use another server:
      dataset.server = 'TRIUMF'
    """
    
    _schema = Schema(Version(1,0), {
        'names': SimpleItem(defvalue = [], typelist=['str'], sequence=1,doc='List of input files with full path to get the file on the server, i.e. "dcap://the/path/thefile'),
        'server': SimpleItem(defvalue = "TRIUMF", doc='Name of the dcache server used'),
    })
    
    _name = 'ND280DCacheDataset'

    _exportmethods = ['get_dataset', 'get_dataset_filenames', 'get_dataset_from_list', 'set_dataset_into_list', 'set_dataset_filenames' ]

    _GUIPrefs = [ { 'attribute' : 'names',  'attribute' :  'server',  'widget' : 'String_List' } ]

    _commandstr = {'TRIUMF' : '#######'}  #TODO: Put this in a config
    _filebasepath = {'TRIUMF' : '#######'} #TODO: Put this in a config
   
    def __init__(self):
        super(ND280DCacheDataset, self).__init__()


    def get_dataset(self,directory,filter=None):
      """Get the list of files in the directory on the dCache server.
      You can provide a wildcard filter such as 'oa_*.root'"""

      if not self.server in self._commandstr.keys():
          logger.error('DCache server %s is unknown.', self.server)
          return

      logger.info('Reading %s ...',directory)

      command = self._commandstr[self.server] % directory
      rawoutput = commands.getoutput(command);
      allfiles = rawoutput.split("\n")

      # TODO: return directory error when curl isn't happy
      #if not os.path.isdir(directory):
      #    logger.error('Path %s is no directory',directory)
      #    return

      fullpath = os.path.join(self._filebasepath[self.server], directory )
      if filter:
          new_names = [ os.path.join(fullpath,name) for name in fnmatch.filter(allfiles,filter) ]
      else:
          new_names = [ os.path.join(fullpath,name) for name in allfiles ]

      self.names.extend( new_names )

      self._setDirty()



from __future__ import absolute_import
# Note: Following stuff must be considered in a GangaRepository:
#
# * lazy loading
# * locking

from .GangaRepository import GangaRepository, RepositoryError
import os
import os.path

import sqlite3

try:
    import cPickle as pickle
except:
    import pickle

import Ganga.Utility.logging
logger = Ganga.Utility.logging.getLogger()


class GangaRepositorySQLite(GangaRepository):

    """GangaRepository SQLite"""

    def startup(self):
        """ Starts an repository and reads in a directory structure."""
        self._load_timestamp = {}
        self._cache_load_timestamp = {}
        self.known_bad_ids = []
        self.root = os.path.join(
            self.registry.location, "0.1", self.registry.name)
        try:
            os.makedirs(self.root)
        except OSError as x:
            pass
        self.con = sqlite3.connect(os.path.join(self.root, "database.db"))
        logger.debug("Connected to ", os.path.join(self.root, "database.db"))
        self.cur = self.con.cursor()
        tables = self.cur.execute(
            "SELECT name FROM sqlite_master WHERE type='table' and name LIKE 'objects'")
        if len(self.cur.fetchall()) == 0:
            self.cur.execute(
                "CREATE TABLE objects (id INTEGER PRIMARY KEY, classname VARCHAR(30), category VARCHAR(20), idx VARCHAR(100), data VARCHAR(1000))")
        self.con.commit()
        self.update_index(verbose=True)

    def shutdown(self):
        """Shutdown the repository. Flushing is done by the Registry
        Raise RepositoryError"""
        self.cur.close()
        self.con.close()

    def update_index(self, id=None, verbose=False):
        """ Update the list of available objects
        Raise RepositoryError"""
        # First locate and load the index files
        logger.debug("updating index...")
        self.cur.execute("SELECT id,classname,category,idx FROM objects")
        for e in self.cur:
            id = int(e[0])
            if e[1] is None:  # deleted object
                continue
            # Locked IDs can be ignored
            # if id in self.sessionlock.locked:
            #    continue
            # Now we treat unlocked IDs
            if id in self.objects:
                obj = self.objects[id]
            else:
                obj = self._make_empty_object_(id, e[2], e[1])
            obj.setNodeIndexCache(pickle.loads(e[3]))
        logger.debug("updated index done")

    def add(self, objs, force_ids=None):
        """ Add the given objects to the repository, forcing the IDs if told to.
        Raise RepositoryError"""
        ids = []
        # assume the ids are already locked by Registry
        if not force_ids is None:
            if not len(objs) == len(force_ids):
                raise RepositoryError(
                    self, "Internal Error: add with different number of objects and force_ids!")
            ids = force_ids
        for i in range(0, len(objs)):
            cls = objs[i]._name
            cat = objs[i]._category
            objs[i].setNodeIndexCache(self.registry.getIndexCache(objs[i]))
            data = pickle.dumps(objs[i].getNodeData()).replace("'", "''")
            idx = pickle.dumps(objs[i].getNodeIndexCache()).replace("'", "''")
            if force_ids is None:
                self.cur.execute("INSERT INTO objects (id,classname,category,idx,data) VALUES (NULL,'%s','%s','%s','%s')" % (
                    cls, cat, idx, data))
            else:
                self.cur.execute("INSERT INTO objects (id,classname,category,idx,data) VALUES (%i,'%s','%s','%s','%s')" % (
                    force_ids[i], cls, cat, idx, data))
            ids.append(self.cur.lastrowid)
            self._internal_setitem__(ids[i], objs[i])
        self.con.commit()
        return ids

    def flush(self, ids):
        for id in ids:
            obj = self.objects[id]
            if obj._name != "EmptyGangaObject":
                obj.setNodeIndexCache(self.registry.getIndexCache(obj))
                data = pickle.dumps(obj.getNodeData()).replace("'", "''")
                idx = pickle.dumps(obj.getNodeIndexCache()).replace("'", "''")
                self.cur.execute(
                    "UPDATE objects SET idx='%s',data='%s' WHERE id=%s" % (idx, data, id))
                # print "flushing id ", id, " backend ", obj.backend._name
        self.con.commit()

    def load(self, ids):
        self.cur.execute("SELECT id,classname,category,data FROM objects WHERE id IN (%s)" % (
            ",".join(map(str, ids))))
        for e in self.cur:
            # print "load: ",e
            id = int(e[0])
            if e[1] is None:  # deleted object
                continue
            if not id in self.objects:
                obj = self._make_empty_object_(id, e[2], e[1])
            else:
                obj = self.objects[id]
            if obj.getNodeData() is None:
                obj.setNodeData( pickle.loads(e[3]) )
            ids.remove(id)
        if len(ids) > 0:
            raise KeyError(ids[0])

    def delete(self, ids):
        for id in ids:
            self.cur.execute(
                "UPDATE objects SET classname=NULL,category=NULL,idx=NULL,data=NULL WHERE id=%s" % (id))
            self._internal_del__(id)
        self.con.commit()

    def lock(self, ids):
        return ids

    def unlock(self, ids):
        return ids

    def get_lock_session(self, id):
        """get_lock_session(id)
        Tries to determine the session that holds the lock on id for information purposes, and return an informative string.
        Returns None on failure
        """
        return ""

    def get_other_sessions(self):
        """get_session_list()
        Tries to determine the other sessions that are active and returns an informative string for each of them.
        """
        return []

    def reap_locks(self):
        """reap_locks() --> True/False
        Remotely clear all foreign locks from the session.
        WARNING: This is not nice.
        Returns True on success, False on error."""
        return False

    def clean(self):
        """clean() --> True/False
        Clear EVERYTHING in this repository, counter, all jobs, etc.
        WARNING: This is not nice."""
        self.shutdown()
        os.unlink(os.path.join(self.root, "database.db"))
        self.startup()



import os
import glob
import tempfile
import copy

from Ganga.Utility.Config import getConfig
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList

from Ganga.GPIDev.Base.Proxy import isType, stripProxy, getName

"""
Checks if the output files of a given job(we are interested in the backend) 
should be postprocessed on the WN, depending on job.backend_output_postprocess dictionary
"""


def outputFilePostProcessingOnWN(job, outputFileClassName):
    backendClassName = getName(job.backend)
    
    backend_output_postprocess = stripProxy(job).getBackendOutputPostprocessDict()
    if backendClassName in backend_output_postprocess:
        if outputFileClassName in backend_output_postprocess[backendClassName]:
            if backend_output_postprocess[backendClassName][outputFileClassName] == 'WN':
                return True

    return False


"""
Checks if the output files of a given job(we are interested in the backend) 
should be postprocessed on the client, depending on job.backend_output_postprocess dictionary
"""


def outputFilePostProcessingOnClient(job, outputFileClassName):
    backendClassName = getName(job.backend)

    backend_output_postprocess = stripProxy(job).getBackendOutputPostprocessDict()
    if backendClassName in backend_output_postprocess:
        if outputFileClassName in backend_output_postprocess[backendClassName]:
            if backend_output_postprocess[backendClassName][outputFileClassName] == 'client':
                return True

    return False

"""
Intented for grid backends where we have to set the outputsandbox patterns for the output file types that have to be processed on the client
"""


def getOutputSandboxPatterns(job):

    outputPatterns = []

    if len(job.outputfiles) > 0:

        outputPatterns.append(getConfig('Output')['PostProcessLocationsFileName'])

        for outputFile in job.outputfiles:

            outputFileClassName = getName(outputFile)

            if outputFilePostProcessingOnClient(job, outputFileClassName) or outputFileClassName == 'LocalFile':
                if outputFile.namePattern not in outputPatterns:
                    if outputFile.compressed:
                        outputPatterns.append('%s.gz' % outputFile.namePattern)
                    else:
                        outputPatterns.append(outputFile.namePattern)

    return outputPatterns

"""
we have to set the inputsandbox patterns for the input files that will be copied from the client, also write the commands for downloading input files from the WN
"""


def getInputFilesPatterns(job):

    tmpDir = tempfile.mkdtemp()

    inputPatterns = []

    # if GangaDataset is used, check if they want the inputfiles transferred
    inputfiles_list = copy.deepcopy(job.inputfiles)
    from Ganga.GPIDev.Lib.Dataset.GangaDataset import GangaDataset
    if not job.subjobs and job.inputdata and isType(job.inputdata, GangaDataset) and job.inputdata.treat_as_inputfiles:
        inputfiles_list += job.inputdata.files

    for inputFile in inputfiles_list:

        inputFileClassName = getName(inputFile)

        if inputFileClassName == 'LocalFile':
            for currentFile in glob.glob(os.path.join(inputFile.localDir, inputFile.namePattern)):
                if currentFile not in inputPatterns:
                    inputPatterns.append(currentFile)

        elif outputFilePostProcessingOnClient(job, inputFileClassName):

            # download in temp dir
            inputFile.localDir = tmpDir
            inputFile.get()

            for currentFile in glob.glob(os.path.join(inputFile.localDir, inputFile.namePattern)):
                if currentFile not in inputPatterns:
                    inputPatterns.append(currentFile)

    return inputPatterns, tmpDir


"""
This should be used from only from Interactive backend
"""


def getOutputSandboxPatternsForInteractive(job):

    patternsToSandbox = [getConfig('Output')['PostProcessLocationsFileName']]
    patternsToZip = []

    for outputFile in job.outputfiles:

        outputFileClassName = getName(outputFile)

        if outputFileClassName == 'LocalFile' or (outputFileClassName != 'LocalFile' and outputFilePostProcessingOnClient(job, outputFileClassName)):
            if outputFile.compressed:
                patternsToSandbox.append('%s.gz' % outputFile.namePattern)
                patternsToZip.append(outputFile.namePattern)
            else:
                patternsToSandbox.append(outputFile.namePattern)

    return (patternsToSandbox, patternsToZip)


"""
This should be used from Local and Batch backend, where there is code on the WN for 
sending the output(optionally compressed before that) to the outputsandbox
"""


def getWNCodeForOutputSandbox(job, files, jobid):

    patternsToSandbox = []
    patternsToZip = []

    if len(job.outputfiles) > 0:

        files.append(getConfig('Output')['PostProcessLocationsFileName'])

        for outputFile in job.outputfiles:

            outputFileClassName = getName(outputFile)

            if outputFileClassName == 'LocalFile' or (outputFileClassName != 'LocalFile' and outputFilePostProcessingOnClient(job, outputFileClassName)):
                patternsToSandbox.append(outputFile.namePattern)

    insertScript = """\n
from Ganga.Utility.files import recursive_copy
import glob
import os

f_to_copy = ###FILES###

for patternToSandbox in ###PATTERNSTOSANDBOX###:
    for currentFile in glob.glob(patternToSandbox):
        f_to_copy.append(currentFile)

filesToZip = []

for patternToZip in ###PATTERNSTOZIP###:
    for currentFile in glob.glob(patternToZip):
        os.system("gzip %s" % currentFile)
        filesToZip.append(currentFile)
            
final_list_to_copy = []

for f in f_to_copy:
    if f in filesToZip:
        final_list_to_copy.append('%s.gz' % f)  
    else:       
        final_list_to_copy.append(f)            

for fn in final_list_to_copy:
    try:
        if not os.path.exists(sharedoutputpath):
            os.makedirs(sharedoutputpath)
        recursive_copy(fn,sharedoutputpath)
    except Exception as x:
        print('ERROR: (job'+###JOBID###+')',x)
"""
    insertScript = insertScript.replace('###FILES###', str(files))
    insertScript = insertScript.replace('###PATTERNSTOSANDBOX###', str(patternsToSandbox))
    insertScript = insertScript.replace('###PATTERNSTOZIP###', str(patternsToZip))
    insertScript = insertScript.replace('###JOBID###', jobid)

    return insertScript


def getWNCodeForDownloadingInputFiles(job, indent):
    """
    Generate the code to be run on the WN to download input files
    """

    from Ganga.GPIDev.Lib.Dataset.GangaDataset import GangaDataset
    if job.inputfiles is None or len(job.inputfiles) == 0 and\
            (not job.inputdata or ((not isType(job.inputdata, GangaDataset)) or\
                not job.inputdata.treat_as_inputfiles )):
        return ""

    insertScript = """\n
"""

    # first, go over any LocalFiles in GangaDatasets to be transferred
    # The LocalFiles in inputfiles have already been dealt with
    if job.inputdata and isType(job.inputdata, GangaDataset) and job.inputdata.treat_as_inputfiles:
        for inputFile in job.inputdata.files:
            inputfileClassName = getName(inputFile)

            if inputfileClassName == "LocalFile":

                # special case for LocalFile
                if getName(job.backend) in ['Localhost', 'Batch', 'LSF', 'Condor', 'PBS']:
                    # create symlink
                    shortScript = """
# create symbolic links for LocalFiles
for f in ###FILELIST###:
   os.symlink(f, os.path.basename(f)) 
"""
                    from Ganga.GPIDev.Lib.File import FileUtils
                    shortScript = FileUtils.indentScript(shortScript, '###INDENT####')

                    insertScript += shortScript

                    insertScript = insertScript.replace('###FILELIST###', "%s" % inputFile.getFilenameList())

    # if GangaDataset is used, check if they want the inputfiles transferred
    inputfiles_list = job.inputfiles
    if job.inputdata and isType(job.inputdata, GangaDataset) and job.inputdata.treat_as_inputfiles:
        inputfiles_list += job.inputdata.files

    for inputFile in inputfiles_list:

        inputfileClassName = getName(inputFile)

        if outputFilePostProcessingOnWN(job, inputfileClassName):
            inputFile.processWildcardMatches()
            if inputFile.subfiles:
                for subfile in inputFile.subfiles:
                    insertScript += subfile.getWNScriptDownloadCommand(indent)
            else:
                insertScript += inputFile.getWNScriptDownloadCommand(indent)

    insertScript = insertScript.replace('###INDENT###', indent)

    return insertScript


def getWNCodeForOutputPostprocessing(job, indent):

    # dict containing the list of outputfiles that need to be processed on the
    # WN for every file type
    outputFilesProcessedOnWN = {}
    patternsToZip = []

    if len(job.outputfiles) == 0:
        return ""
    else:
        for outputFile in job.outputfiles:

            outputfileClassName = getName(outputFile)
            backendClassName = getName(job.backend)

            if outputFile.compressed:
                if outputfileClassName == 'LocalFile' and backendClassName not in ['Localhost', 'LSF', 'Interactive']:
                    patternsToZip.append(outputFile.namePattern)
                elif outputfileClassName != 'LocalFile' and outputFilePostProcessingOnWN(job, outputfileClassName):
                    patternsToZip.append(outputFile.namePattern)
                elif outputfileClassName != 'LocalFile' and outputFilePostProcessingOnClient(job, outputfileClassName) and backendClassName not in ['Localhost', 'LSF', 'Interactive']:
                    patternsToZip.append(outputFile.namePattern)

            if outputfileClassName not in outputFilesProcessedOnWN.keys():
                outputFilesProcessedOnWN[outputfileClassName] = []

            if outputFilePostProcessingOnWN(job, outputfileClassName):
                outputFilesProcessedOnWN[
                    outputfileClassName].append(outputFile)

    shortScript = """\n
import os, glob
for patternToZip in ###PATTERNSTOZIP###:
    for currentFile in glob.glob(os.path.join(os.getcwd(),patternToZip)):
        os.system("gzip %s" % currentFile)

postprocesslocations = file(os.path.join(os.getcwd(), '###POSTPROCESSLOCATIONSFILENAME###'), 'w')  
"""

    from Ganga.GPIDev.Lib.File import FileUtils
    shortScript = FileUtils.indentScript(shortScript, '###INDENT###')

    insertScript = shortScript

    insertScript = insertScript.replace('###PATTERNSTOZIP###', str(patternsToZip))
    insertScript = insertScript.replace('###POSTPROCESSLOCATIONSFILENAME###', getConfig('Output')['PostProcessLocationsFileName'])

    for outputFileName in outputFilesProcessedOnWN.keys():

        if len(outputFilesProcessedOnWN[outputFileName]) > 0:

            insertScript += outputFilesProcessedOnWN[outputFileName][0].getWNInjectedScript(outputFilesProcessedOnWN[outputFileName], indent, patternsToZip, 'postprocesslocations')

    insertScript += """\n
###INDENT###postprocesslocations.close()
"""
    insertScript = insertScript.replace('###INDENT###', indent)

    return insertScript

import re

wildcardregex = re.compile('[*?\[\]]')


def iexpandWildCards(filelist):
    for f in filelist:
        if wildcardregex.search(f.namePattern):
            f.processWildcardMatches()
            for subfile in f.subfiles:
                yield subfile
        else:
            yield f


def expandWildCards(filelist):
    """

    """
    l = GangaList()
    l.extend(iexpandWildCards(filelist))
    return l


def getWNCodeForInputdataListCreation(job, indent):
    """generate the code to create ths inputdata list on the worker node"""
    insertScript = """\n
###INDENT###open("__GangaInputData.txt__", "w").write( "\\n".join( ###FILELIST### ) )
"""

    insertScript = insertScript.replace('###INDENT###', indent)

    if job.inputdata and hasattr(job.inputdata, "getFilenameList"):
        insertScript = insertScript.replace(
            '###FILELIST###', "%s" % job.inputdata.getFilenameList())
    else:
        insertScript = insertScript.replace('###FILELIST###', "[]")

    return insertScript




import fnmatch
import copy
from Ganga.Core.GangaRepository.Registry import Registry, RegistryKeyError
from Ganga.GPIDev.Lib.Registry.RegistrySlice import RegistrySlice
from Ganga.GPIDev.Lib.Registry.RegistrySliceProxy import RegistrySliceProxy, _wrap
from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.Utility.logging import getLogger
logger = getLogger()


class TransientRegistry(Registry):

    def __init__(self, name, filebase, doc, file_ext='tpl', pickle_files=False, dirty_flush_counter=10, update_index_time=30):

        super(TransientRegistry, self).__init__(name,
                                                doc,
                                                dirty_flush_counter,
                                                update_index_time)
        self.type = "ImmutableTransient"
        self.location = filebase
        self.file_ext = file_ext
        self.pickle_files = pickle_files
        self._needs_metadata = False

        self.stored_slice = TransientRegistrySlice(self.name)
        self.stored_slice.objects = self
        self.stored_slice.add = self.add
        self.stored_proxy = TransientRegistrySliceProxy(self.stored_proxy)

    # def startup(self):
        # Note call the base class setup as dont want
        # metadata which JobRegistry forces on us
        #super(TransientRegistry, self).startup()
        # Registry.startup(self)

    def _getName(self, obj):
        return obj.name

    def add(self, obj, name):
        """
        Add an object to the registry
        """
        o = copy.deepcopy(stripProxy(obj))
        #o = stripProxy(obj)
        o.name = name
        #o._registry = self
        #o._registry_id = -1
        # print stripProxy(obj), o
        super(TransientRegistry, self)._add(o)

    def getSlice(self):
        return self.stored_slice

    def getProxy(self):
        return self.stored_proxy


class TransientRegistrySlice(RegistrySlice):

    def __init__(self, name):
        super(TransientRegistrySlice, self).__init__(
            name, display_prefix="box")
        from Ganga.Utility.ColourText import Foreground, Background, Effects
        fg = Foreground()
        fx = Effects()
        bg = Background()
        self.fx = fx
        self.name = 'box'  # needed to ensure that select works properly
        self.status_colours = {'default': fx.normal,
                               'JobTemplate': fg.orange,
                               'Task': fg.green,
                               'Job': fg.blue}
        self._display_columns_functions["id"] = lambda obj: obj.id
        self._display_columns_functions["type"] = lambda obj: obj._name
        self._display_columns_functions["name"] = lambda obj: obj.name
        self._proxyClass = TransientRegistrySliceProxy

    def _getColour(self, obj):
        try:
            return self.status_colours.get(getName(obj), self.fx.normal)
        except Exception as err:
            return self.status_colours['default']

    def __call__(self, id):
        """
        Retrieve an object by id.
        """
        if isinstance(id, str):
            if id.isdigit():
                id = int(id)
            else:
                matches = [
                    o for o in self.objects if fnmatch.fnmatch(o.name, id)]
                if len(matches) > 1:
                    logger.error(
                        'Multiple Matches: Wildcards are allowed for ease of matching, however')
                    logger.error(
                        '                  to keep a uniform response only one item may be matched.')
                    logger.error(
                        '                  If you wanted a slice, please use the select method')
                    raise RegistryKeyError(
                        "Multiple matches for id='%s':%s" % (id, str(map(lambda x: x.name, matches))))
                if len(matches) < 1:
                    return _wrap(TransientRegistrySlice(self.name))
                return matches[0]
        try:
            return self.objects[id]
        except KeyError:
            raise RegistryKeyError('Object id=%d not found' % id)


class TransientRegistrySliceProxy(RegistrySliceProxy):

    def __init__(self, impl):
        super(TransientRegistrySliceProxy, self).__init__(impl)

    def __call__(self, x):
        """
        Access individual object. Examples:
        """
        return _wrap(stripProxy(self).__call__(x))

    def __getitem__(self, x):
        """
        Get an item by positional index. Examples:
        """
        return _wrap(stripProxy(self).__getitem__(x))

    def __getslice__(self, i1, i2):
        """
        Get a slice. Examples:
        """
        return _wrap(stripProxy(self).__getslice__(i1, i2))



from Ganga.GPIDev.Lib.File.Configure import getSharedPath

from Ganga.GPIDev.Lib.File.File import File
from Ganga.GPIDev.Lib.File.File import ShareDir
from Ganga.GPIDev.Lib.File.FileBuffer import FileBuffer

from Ganga.GPIDev.Lib.File.LocalFile import LocalFile
from Ganga.GPIDev.Lib.File.MassStorageFile import MassStorageFile
from Ganga.GPIDev.Lib.File.LCGSEFile import LCGSEFile
from Ganga.GPIDev.Lib.File.SandboxFile import SandboxFile
from Ganga.GPIDev.Lib.File.GoogleFile import GoogleFile

import Ganga.Utility.logging

from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.GPIDev.Base.Filters import allComponentFilters
from Ganga.Utility.Config import getConfig, ConfigError


import fnmatch

logger = Ganga.Utility.logging.getLogger()


def getFileConfigKeys():
    keys = getConfig('Output').options.keys()
    keys.remove('PostProcessLocationsFileName')
    keys.remove('ForbidLegacyInput')
    keys.remove('ForbidLegacyOutput')
    keys.remove('AutoRemoveFilesWithJob')
    keys.remove('AutoRemoveFileTypes')
    keys.remove('FailJobIfNoOutputMatched')
    return keys


def decodeExtensionKeys():

    outputfilesConfig = {}

    keys = getFileConfigKeys()

    for key in keys:
        try:
            outputFilePatterns = []

            for configEntry in getConfig('Output')[key]['fileExtensions']:
                outputFilePatterns.append(configEntry)

            outputfilesConfig[key] = outputFilePatterns

        except ConfigError as err:
            logger.debug("ConfigureError: %s" % str(err))
            pass

    return outputfilesConfig


def findOutputFileTypeByFileName(filename):

    matchKeys = []

    outputfilesConfig = decodeExtensionKeys()

    for key in outputfilesConfig.keys():
        for filePattern in outputfilesConfig[key]:
            if fnmatch.fnmatch(filename, filePattern):
                matchKeys.append(key)

    if len(matchKeys) == 1:
        #logger.debug("File name pattern %s matched %s, assigning to %s" % (
        #    filename, str(matchKeys), matchKeys[-1]))
        return matchKeys[-1]
    elif len(matchKeys) > 1:
        #logger.warning("file name pattern %s matched %s, assigning to %s" % (
        #    filename, str(matchKeys), matchKeys[-1]))
        return matchKeys[-1]
    else:
        #logger.debug("File name pattern %s is not matched" % filename)
        return None


def string_file_shortcut(v, item):
    if isinstance(v, str):
        # use proxy class to enable all user conversions on the value itself
        # but return the implementation object (not proxy)
        key = findOutputFileTypeByFileName(v)
        if key is not None:
            if key == 'MassStorageFile':
                from .MassStorageFile import MassStorageFile
                return stripProxy(MassStorageFile._proxyClass(v))
            elif key == 'LCGSEFile':
                from .LCGSEFile import LCGSEFile
                return stripProxy(LCGSEFile._proxyClass(v))
            elif key == 'DiracFile':
                try:
                    from GangaDirac.Lib.Files.DiracFile import DiracFile
                    return stripProxy(DiracFile._proxyClass(v))
                except:
                    Ganga.Utility.logging.log_unknown_exception()
                    pass

        return stripProxy(LocalFile._proxyClass(v))

    return None

allComponentFilters['gangafiles'] = string_file_shortcut




"""File emailer IAction implementation.

The FileEmailer class can be configured to email text and/or html files.
 
"""

from GangaRobot.Framework.Action import IAction
from GangaRobot.Framework import Utility
from Ganga.Utility.logging import getLogger
from email.MIMEText import MIMEText
from email.MIMEMultipart import MIMEMultipart
from smtplib import SMTP

logger = getLogger()

class FileEmailer(IAction):
    
    """File emailer IAction implementation.
    
    A configurable action which emails text and/or html files to one or more
    recipients. See execute() for details.
    
    """

    def execute(self, runid):
        """Send files as an email.
        
        Keyword arguments:
        runid -- A UTC ID string which identifies the run.
        
        The following parameters are loaded from the Robot configuration:
        FileEmailer_Host e.g. localhost or localhost:25
        FileEmailer_Type e.g. text or html
        FileEmailer_From e.g. sender@domain.org
        FileEmailer_Recipients e.g. recipient1@domain.org, recipient2@domain.org
        FileEmailer_Subject e.g. Report ${runid}.
        FileEmailer_TextFile e.g. ~/gangadir/robot/report/${runid}.txt
        FileEmailer_HtmlFile e.g. ~/gangadir/robot/report/${runid}.html
        
        If Recipients are not specified then no email is sent.
        In Subject, TextFile and HtmlFile the token ${runid} is replaced by the
        runid argument.
        If Type is text, then TextFile is sent.
        If Type is html, then HtmlFile is sent, or if TextFile is also specified
        then a multipart message is sent containing TextFile and HtmlFile.
        
        """
        # get configuration properties
        host = self.getoption('FileEmailer_Host')
        type = self.getoption('FileEmailer_Type')
        from_ = self.getoption('FileEmailer_From')
        # extract recipients ignoring blank entries
        recipients = [recipient.strip() for recipient in \
                      self.getoption('FileEmailer_Recipients').split(',') \
                      if recipient.strip()]
        subject = Utility.expand(self.getoption('FileEmailer_Subject'), runid = runid)
        textfilename = Utility.expand(self.getoption('FileEmailer_TextFile'), runid = runid)
        htmlfilename = Utility.expand(self.getoption('FileEmailer_HtmlFile'), runid = runid)
        
        if not recipients:
            logger.warn('No recipients specified. Email will not be sent.')
            return
        
        logger.info('Emailing files to %s.', recipients)

        # build message
        if type == 'html':
            msg = self._gethtmlmsg(textfilename, htmlfilename)
        else:
            msg = self._gettextmsg(textfilename)
        msg['Subject'] = subject
        msg['From'] = from_
        msg['To'] = ', '.join(recipients)

        # send message
        session = SMTP()
        try:
            session.connect(host)
            session.sendmail(from_, recipients, msg.as_string())
            session.quit()
        finally:
            session.close()

        logger.info('Files emailed.')
            
    def _gettextmsg(self, textfilename):
        text = Utility.readfile(textfilename)
        textmsg = MIMEText(text)
        return textmsg
    
    def _gethtmlmsg(self, textfilename, htmlfilename):
        html = Utility.readfile(htmlfilename)
        htmlmsg = MIMEText(html, 'html')
        if textfilename:
            textmsg = self._gettextmsg(textfilename)
            multimsg = MIMEMultipart('alternative')
            multimsg.attach(textmsg)
            multimsg.attach(htmlmsg)
            return multimsg
        else:
            return htmlmsg
        



##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: __init__.py,v 1.1 2008-07-17 16:40:55 moscicki Exp $
##########################################################################

# File: Persistency/__init__.py
# Author: K. Harrison
# Created: 051020
# Last modified: 051021

"""Initialisation file for the Persistency package,
   containing functions for exporting and loading Ganga objects"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "21 October 2005"
__version__ = "1.0"

import Ganga.GPI
from Ganga.Utility.Runtime import getScriptPath, getSearchPath
from Ganga.GPIDev.Base.Proxy import stripProxy, isType
from Ganga.GPIDev.Lib.Registry.RegistrySlice import RegistrySlice
from Ganga.GPIDev.Lib.Registry.RegistrySliceProxy import RegistrySliceProxy
from Ganga.GPIDev.Lib.GangaList.GangaList import GangaList
import Ganga.Utility.logging
import os
import sys
import time

logger = Ganga.Utility.logging.getLogger()

# This is exported at bootstrap to the GPI


def export(item=None, filename="", mode="w"):
    return stripped_export(stripProxy(item), filename, mode)

# Now we can use it internally without having to wrap a proxy


def stripped_export(item=None, filename="", mode="w"):
    """Function to export Ganga objects to a file

       Arguments:
          item       - Ganga object(s) [default None] to be exported
                       => item may be any of following:
                          a single Ganga object;
                          a list or tuple of Ganga objects;
                          a repository object (jobs or templates)
          filename   - String [default ''] giving relative or absolute path
                       to file where object definitions are to be exported
          mode       - String [default 'w'] giving mode in which file
                       specified by 'filename' is to be accessed:
                          'w' : write to new file
                          'a' : append to existing file

       Ganga objects saved to a file with the 'export' command can
       be loaded with the 'load' command

       Return value: True if Ganga object(s) successfully written to file,
                     or False otherwise
    """

    returnValue = False

    if not item or not filename:
        logger.info("Usage:")
        logger.info("export( <GangaObject>, '<filename>', [ '<mode>' ] )")
        logger.info("See also: 'help( export )'")
        return returnValue

    modeDict = {
        "w": "write to new file",
        "a": "append to existing file"}

    if mode not in modeDict.keys():
        logger.info("'mode' must be one of:")
        for key in modeDict.keys():
            logger.info("   '%s' - %s" % (key, modeDict[key]))
        logger.info("No object saved")
        return returnValue

    filepath = fullpath(filename)
    try:
        outFile = open(filepath, mode)
    except IOError:
        logger.error("Unable to open file '%s' for writing" % filename)
        logger.error("No object saved")
        return returnValue

    if isinstance(item, list):
        objectList = [stripProxy(element) for element in item]
    elif isinstance(item, tuple):
        objectList = [stripProxy(element) for element in item]
    elif isType(item, RegistrySliceProxy) or isType(item, RegistrySlice):
        objectList = [stripProxy(element) for element in item]
    elif isType(item, GangaList):
        objectList = [stripProxy(element) for element in item]
    else:
        objectList = [item]

    if mode == "w":
        lineList = [
            "#Ganga# File created by Ganga - %s\n" % (time.strftime("%c")),
            "#Ganga#\n",
            "#Ganga# Object properties may be freely edited before reloading into Ganga\n",
            "#Ganga#\n",
            "#Ganga# Lines beginning #Ganga# are used to divide object definitions,\n",
            "#Ganga# and must not be deleted\n",
            "\n"]
        outFile.writelines(lineList)

    nObject = 0
    for this_object in objectList:
        try:
            name = this_object._name
            category = this_object._category
            outFile.write("#Ganga# %s object (category: %s)\n" % (name, category))
            this_object.printTree(outFile, "copyable")
            nObject = nObject + 1
        except AttributeError as err:
            logger.info("Unable to save item - not a GangaObject")
            logger.info("Problem item: %s" % (repr(object)))
            logger.warning("Error: %s" % str(err))
            raise err
        except Exception as err:
            logger.error("Unknown export error!")
            raise err

    if nObject:
        returnValue = True
    else:
        logger.warning("No objects saved to file '%s'" % filename)

    outFile.close()

    return returnValue


def fullpath(filename=""):
    """Function to determine the full path corresponding to a given filename,
       expanding environment variables and tilda as necessary

       Argument:
          filename - String [default ''] giving the name of a file

       Return value: String giving absolute path to file
    """
    filepath = os.path.abspath\
        (os.path.expandvars(os.path.expanduser(filename)))
    return filepath


def load(filename="", returnList=True):
    """Function to load previously exported Ganga objects

       Arguments:
          filename   - String [default ''] giving path to a file containing
                       definitions of Ganga objects, such as produced with 
                       the 'export' function
                       => The path can be absolute, relative, or relative
                          to a directory in the search path LOAD_SCRIPTS,
                          defined in [Configuration] section of
                          Ganga configuration
          returnList - Switch [default True] defining the return type:
                       => True  : List of loaded objects is returned
                       => False : None returned - job and template objects
                                  stored in job repository

       Return value: List of Ganga objects or None, as determined by value
                     of argument returnList
    """

    #logger.info("Loading: %s" % str(filename))

    if returnList:
        returnValue = []
    else:
        returnValue = None

    if not filename:
        logger.info("Usage:")
        logger.info("load( '<filename>', [ <returnList> ] )")
        logger.info("See also: 'help( load )'")
        return returnValue

    searchPath = getSearchPath("LOAD_PATH")
    filepath = getScriptPath(filename, searchPath)
    try:
        with open(filepath) as inFile:
            lineList = []
            for line in inFile:
                if (line.strip().startswith("#Ganga#")):
                    lineList.append("#Ganga#")
                else:
                    lineList.append(line)
            itemList = ("".join(lineList)).split("#Ganga#")
    except IOError:
        logger.error("Unable to open file %s" % (str(filename)))
        logger.error("No objects loaded")
        return returnValue

    objectList = []
    for item in itemList:
        item = item.strip()
        if item:
            try:
                from Ganga.GPIDev.Base.Proxy import getProxyInterface
                this_object = eval(str(item), getProxyInterface().__dict__)
                objectList.append(this_object)
            except NameError as x:
                logger.exception(x)
                logger.warning("Unable to load object with definition %s" % item)
                logger.warning("Required plug-ins may not be available")

    if not objectList:
        logger.warning("Unable to load any objects from file %s" % filename)

    if returnList:
        returnValue = objectList

    return returnValue




##################################################
# retrieve and cache monitoring classes by name
_mon_classes = {}

from Ganga.Core.exceptions import GangaException
from Ganga.GPIDev.Adapters.IMonitoringService import IMonitoringService
from Ganga.GPIDev.MonitoringServices.Composite import CompositeMonitoringService

from Ganga.Utility.Config import getConfig

from Ganga.Utility.logging import getLogger
logger = getLogger()

c = getConfig('MonitoringServices')


class MonitoringServiceError(GangaException):

    def __init__(self):
        super(MonitoringServiceError, self).__init__()


def getMonitoringClass(mclassname):
    """
    Return the class object based on the class name string provided as input
    If the class object is already available in cache the saved value is returned
    """
    if mclassname in _mon_classes:
        return _mon_classes[mclassname]
    else:
        try:
            logger.debug("getMonClass: %s" % str(mclassname))
            classpath = mclassname.split('.')
            classname = classpath.pop()
            modname = '.'.join(classpath)
            logger.debug("modname: %s" % modname)
            monitoring_module = __import__(modname, globals(), locals(), [classname])
            logger.debug("monitoring_module: %s" % str(monitoring_module))
            monitoring_class = vars(monitoring_module)[classname]

            try:
                if not issubclass(monitoring_class, IMonitoringService):
                    raise MonitoringServiceError('%s is not IMonitoringService subclass while loading %s' % (classname, mclassname))
            except TypeError as err:
                logger.debug("TypeError1: %s" % str(err))
                raise MonitoringServiceError('%s (%s) is not IMonitoringService subclass while loading %s' % (classname, str(type(monitoring_class)), mclassname))

            # store the modname as class variable
            monitoring_class._mod_name = modname

            _mon_classes[mclassname] = monitoring_class
            return monitoring_class
        except ImportError as err:
            logger.debug("ImportError: %s" % str(err))
            raise MonitoringServiceError('%s while loading %s' % (str(err), mclassname))
        except KeyError as err:
            logger.debug("KeyError %s" % str(err))
            raise MonitoringServiceError('class %s not found while loading %s' % (classname, mclassname))


def findMonitoringClassesName(job):
    """
    Return a comma separted list of class names for 
    a gived job based on its backend and application names.
    """

    from Ganga.Utility.Config import getConfig, ConfigError
    mc = getConfig('MonitoringServices')

    def _getMonClasses(option):
        """
        for a given app,backend pair return the monitoring classes or an empty string
        if it's None or not defined
        """
        if option in mc:
            # try:
            monClasses = mc[option]
            # bug #65444: never return None
            if monClasses:
                return monClasses
        # except (ConfigError,KeyError):
        #   pass
        return ''

    # we try config parameters:
    #  - Application/Backend
    #  - Application/*
    #  - */Backend
    #  - */*
    from Ganga.GPIDev.Base.Proxy import getName
    applicationName = getName(job.application)
    backendName = getName(job.backend)

    allclasses = []
    for configParam in [applicationName + '/' + backendName,
                        applicationName + '/*',
                        '*/' + backendName,
                        '*/*']:
        allclasses += _getMonClasses(configParam).split(",")

    # remove double entries
    uniqueclasses = []
    for x in allclasses:
        if len(x) > 0 and x not in uniqueclasses:
            uniqueclasses += [x]

    return ",".join(uniqueclasses)


def getMonitoringObject(job):
    """
    Composite pattern: 
     return a wrapped object implementing the IMonitoringService which contains a 
     list of IMonitoringServices inside and delegating the interface methods to each of them
    """
    # read from configuration
    names = [name.strip() for name in findMonitoringClassesName(job).split(',') if name.strip()]
    # get classes, jobs and configs
    monClasses = []
    for  name in names:
        try:
            this_class = getMonitoringClass(name)
        except MonitoringServiceError as err:
            logger.debug("Error with: %s" % str(err))
            this_class = None
        if this_class is not None:
            monClasses.append(this_class)
    jobs = [job] * len(monClasses)
    configs = [monClass.getConfig() for monClass in monClasses]
    return CompositeMonitoringService(monClasses, jobs, configs)




#!/usr/bin/env python
'''Script to run Gaudi application'''

from os import curdir, system, environ, pathsep, sep, getcwd
from os.path import join
import sys

def prependEnv(key, value):
    if key in environ: value += (pathsep + environ[key])
    environ[key] = value

# Main
if __name__ == '__main__':

    prependEnv('LD_LIBRARY_PATH', getcwd() + '/lib')
    prependEnv('PYTHONPATH', getcwd() + '/InstallArea/python')
    prependEnv('PYTHONPATH', getcwd() + '/InstallArea/###PLATFORM###/python')

    rc = system('''###COMMAND###''')/256

    ###XMLSUMMARYPARSING###

    ###OUTPUTFILESINJECTEDCODE###

    sys.exit(rc)




"""
Core package defines the fundamental subsystems of Ganga Architecture.
Subsystems are autonomous components (such as a remote services) which may be independetly deployed.
Subsystems may also be created as local objects in the Ganga Client process.
"""
from __future__ import absolute_import

import time

from .exceptions import GangaException, ApplicationConfigurationError, \
    BackendError, RepositoryError, BulkOperationRepositoryError, \
    IncompleteJobSubmissionError, IncompleteKillError, JobManagerError, \
    GangaAttributeError, GangaValueError, ProtectedAttributeError, \
    ReadOnlyObjectError, TypeMismatchError, SchemaError, ApplicationPrepareError, \
    GangaIOError, SplitterError

monitoring_component = None


def set_autostart_policy(interactive_session):
    """
    Change the default value of autostart of the monitoring, depending if the session is interactive or batch.
    The autostart value may be overriden in the config file, so warn if it differs from the default.
    This function should be called
    """
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import config

# internal helper variable for interactive shutdown
t_last = None


def start_jobregistry_monitor(reg_slice):
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import JobRegistry_Monitor
    # start the monitoring loop
    global monitoring_component
    monitoring_component = JobRegistry_Monitor(reg_slice)
    monitoring_component.start()


def bootstrap(reg_slice, interactive_session, my_interface=None):
    """
    Create local subsystems. In the future this procedure should be enhanced to connect to remote subsystems.
    FIXME: this procedure should be moved to the Runtime package.

    This function will change the default value of autostart of the monitoring, depending if the session is interactive or batch.
    The autostart value may be overriden in the config file, so warn if it differs from the default.
    """
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import JobRegistry_Monitor, config
    from Ganga.Utility.logging import getLogger

    logger = getLogger()

    from Ganga.Core.GangaThread import GangaThreadPool

    # start the internal services coordinator
    from Ganga.Core.InternalServices import Coordinator
    Coordinator.bootstrap()

    # backend-specific setup (e.g. Remote: setup any remote ssh pipes)
    # for j in reg_slice:
    #    if hasattr(j,'status') and j.status in ['submitted','running']:
    #        if hasattr(j,'backend'): # protect: EmptyGangaObject does not have backend either
    #            if hasattr(j.backend,'setup'): # protect: EmptyGangaObject does not have setup() method
    #                j.backend.setup()

    start_jobregistry_monitor(reg_slice)

    # register the MC shutdown hook

    change_atexitPolicy(interactive_session)

    # export to GPI moved to Runtime bootstrap

    autostart_default = interactive_session
    config.overrideDefaultValue('autostart', bool(autostart_default))

    if config['autostart'] is not autostart_default:
        msg = 'monitoring loop %s (the default setting for %s session is %s)'
        val = {True: ('enabled', 'batch', 'disabled'),
               False: ('disabled', 'interactive', 'enabled')}
        logger.warning(msg % val[config['autostart']])

    if config['autostart']:
        monitoring_component.enableMonitoring()

    if not my_interface:
        import Ganga.GPI
        my_interface = Ganga.GPI
    from Ganga.Runtime.GPIexport import exportToInterface
    exportToInterface(my_interface, 'runMonitoring', monitoring_component.runMonitoring, 'Functions')
        

def should_wait_interactive_cb(t_total, critical_thread_ids, non_critical_thread_ids):
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import config
    global t_last
    if t_last is None:
        t_last = -time.time()
    # if there are critical threads then prompt user or wait depending on
    # configuration
    if critical_thread_ids:
        if ((t_last < 0 and time.time() + t_last > config['forced_shutdown_first_prompt_time']) or
                (t_last > 0 and time.time() - t_last > config['forced_shutdown_prompt_time'])):
            msg = """Job status update or output download still in progress (shutdown not completed after %d seconds).
%d background thread(s) still running: %s.
Do you want to force the exit (y/[n])? """ % (t_total, len(critical_thread_ids), critical_thread_ids)
            resp = raw_input(msg)
            t_last = time.time()
            return resp.lower() != 'y'
        else:
            return True
    # if there are non-critical threads then wait or shutdown depending on
    # configuration
    elif non_critical_thread_ids:
        if t_total < config['forced_shutdown_first_prompt_time']:
            return True
        else:
            return False
    # if there are no threads then shutdown
    else:
        return False


def should_wait_batch_cb(t_total, critical_thread_ids, non_critical_thread_ids):
    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import config
    from Ganga.Utility.logging import getLogger
    # if there are critical threads then wait or shutdown depending on
    # configuration
    if critical_thread_ids:
        if t_total < config['forced_shutdown_timeout']:
            return True
        else:
            getLogger().warning('Shutdown was forced after waiting for %d seconds for background activities to finish\
(monitoring, output download, etc). This may result in some jobs being corrupted.', t_total)
            return False
    # if there are non-critical threads then wait or shutdown depending on
    # configuration
    elif non_critical_thread_ids:
        if t_total < config['forced_shutdown_first_prompt_time']:
            return True
        else:
            return False
    # if there are no threads then shutdown
    else:
        return False

at_exit_should_wait_cb = None

current_shutdown_policy = None

def change_atexitPolicy(interactive_session=True, new_policy=None):

    from Ganga.Core.MonitoringComponent.Local_GangaMC_Service import config

    if new_policy is None:
        # select the shutdown method based on configuration and/or session type
        forced_shutdown_policy = config['forced_shutdown_policy']
    else:
        forced_shutdown_policy = new_policy

    if forced_shutdown_policy == 'interactive':
        should_wait_cb = should_wait_interactive_cb
    else:
        if forced_shutdown_policy == 'batch':
            should_wait_cb = should_wait_batch_cb
        else:
            if interactive_session:
                should_wait_cb = should_wait_interactive_cb
            else:
                should_wait_cb = should_wait_batch_cb

    global at_exit_should_wait_cb
    if at_exit_should_wait_cb is None:
        at_exit_should_wait_cb = should_wait_cb
        import atexit
        from Ganga.Core.GangaThread import GangaThreadPool
        # create generic Ganga thread pool
        thread_pool = GangaThreadPool.getInstance()
        atexit.register(thread_pool.shutdown, should_wait_cb=at_exit_should_wait_cb)
    else:
        at_exit_should_wait_cb = should_wait_cb


    global current_shutdown_policy
    current_shutdown_policy = forced_shutdown_policy

def getCurrentShutdownPolicy():
    global current_shutdown_policy
    if current_shutdown_policy is None:
        current_shutdown_policy = config['forced_shutdown_policy']
        return config['forced_shutdown_policy']
    return current_shutdown_policy




from copy import deepcopy

from Ganga.GPIDev.Base.Proxy import stripProxy
from Ganga.GPIDev.Base.Proxy import isType
from Ganga.GPIDev.Schema import Schema, Version
from Ganga.Core.exceptions import GangaException
from Ganga.Core.exceptions import GangaAttributeError
from Ganga.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
import time
from GangaLHCb.Lib.LHCbDataset.BKQuery import BKQuery
from Ganga.GPIDev.Lib.Tasks import ITask
from GangaLHCb.Lib.Tasks.LHCbTransform import LHCbTransform
from Ganga.Utility.logging import getLogger

logger = getLogger()

########################################################################


class LHCbTask(ITask):

    """LHCb add-ons for the Task framework"""
    _schema = Schema(Version(1, 0), dict(ITask._schema.datadict.items() + {
    }.items()))

    _category = 'tasks'
    _name = 'LHCbTask'
    _exportmethods = ITask._exportmethods + ['addQuery', 'updateQuery', 'removeUnusedData', 'cleanTask']

    _tasktype = "ITask"

    default_registry = "tasks"

    def cleanTask(self):
        """Delete unused data and then remove all unused jobs"""
        self.removeUnusedData()
        self.removeUnusedJobs()

    def removeUnusedData(self):
        """Remove any output data from orphaned jobs"""
        for trf in self.transforms:
            if hasattr(trf, "removeUnusedData"):
                trf.removeUnusedData()

    def addQuery(self, transform, bkQuery, associate=True):
        """Allows the user to add multiple transforms corresponding to the list of
        BKQuery type objects given in the second argument. The first argument
        is a transform object to use as the basis for the creation of further
        transforms."""
        if not isinstance(transform, LHCbTransform):
            raise GangaException(
                None, 'First argument must be an LHCbTransform object to use as the basis for establishing the new transforms')

        # Check if the template transform is associated with the Task
        try:
            self.transforms.index(transform)
        except:
            if associate:
                logger.info(
                    'The transform is not associated with this Task, doing so now.')
                self.appendTransform(transform)

        # Check if the BKQuery input is correct and append/update
        if type(bkQuery) is not list:
            bkQuery = [bkQuery]
        for bk in bkQuery:
            if not isType(bk, BKQuery):
                raise GangaAttributeError(
                    None, 'LHCbTransform expects a BKQuery object or list of BKQuery objects passed to the addQuery method')
            if len(transform.queries) != 0:  # If template has no query itself
                logger.info('Attaching query to transform')
                transform.addQuery(stripProxy(bk))
            else:  # Duplicate from template
                logger.info('Duplicating transform to add new query.')
                tr = deepcopy(transform)
                tr.addQuery(stripProxy(bk))
                self.appendTransform(tr)

    def appendTransform(self, transform):
        """Append a transform to this task. This method also performs an update() on
        the transform once successfully appended."""
        r = super(LHCbTask, self).appendTransform(transform)
        return r

    def updateQuery(self, resubmit=False):
        """Update the dataset information of all attached transforms. This will
        include any new data in the processing or re-run jobs that have data which
        has been removed."""
        # Tried to use multithreading, better to check the tasksregistry class
        # Also tried multiprocessing but bottleneck at server.
        for t in self.transforms:
            try:
                t.updateQuery(resubmit)
            except GangaException as e:
                logger.warning(e.__str__())
                continue

        # update the status of the Task in case we're started running again
        self.updateStatus()



def getEnvironment(c):
    import PACKAGE
    PACKAGE.standardSetup()
    return {}
    

def loadPlugins(c):
    pass

from GangaTutorial.Lib.primes.primes import check_prime_job, split_prime_job

from Ganga.Runtime.GPIexport import exportToGPI

from Ganga.Utility.logging import getLogger
logger = getLogger(modulename=True)

exportToGPI("check_prime_job",check_prime_job,"Functions")
exportToGPI("split_prime_job",split_prime_job,"Functions")

import os, GangaTutorial
TUTDIR = os.path.dirname(GangaTutorial.__file__)
exportToGPI("TUTDIR",TUTDIR,"Objects")
del os,GangaTutorial

logger.debug("*** Ganga Tutorial Loaded OK ***")



from GangaTest.Framework.tests import GangaGPITestCase
#from unittest import TestCase

from GangaRobot.Framework import Utility

class TestUtility(GangaGPITestCase):

    """Tests of utility methods.
    
    N.B. This test would work as a standard TestCase.
    """
    
    def test_utcid(self):
        """Test utcid() returns a string value when no parameter."""
        utcid = Utility.utcid()
        assert isinstance(utcid, str), 'utcid is not a string'
        
    def test_utctime(self):
        """Test utctime() returns a string value when no parameter."""
        utctime = Utility.utctime()
        assert isinstance(utctime, str), 'utctime is not a string'
        
    def test_utcid_from_utctime(self):
        """Test utcid() returns utcid corresponding to utctime parameter."""
        utctime = Utility.utctime()
        utcid = Utility.utcid(utctime)
        assert isinstance(utcid, str), 'utcid is not a string'
        assert utctime == Utility.utctime(utcid), 'utctime -> utcid -> utctime conversion fails'

    def test_utctime_from_utcid(self):
        """Test utctime() returns utctime corresponding to utcid parameter."""
        utcid = Utility.utcid()
        utctime = Utility.utctime(utcid)
        assert isinstance(utctime, str), 'utctime is not a string'
        assert utcid == Utility.utcid(utctime), 'utcid -> utctime -> utcid conversion fails'

    def test_expand_no_replacements(self):
        """Test expand() does not replace tokens in text when no replacements parameter."""
        text = 'The following token ${runid} should not be replaced.'
        expected = text
        actual = Utility.expand(text)
        assert expected == actual, 'the text was modified'

    def test_expand_single_occurrence(self):
        """Test expand() replaces token in text."""
        text = 'The following token ${mytoken1} should be replaced.'
        expected = 'The following token 1111 should be replaced.'
        actual = Utility.expand(text, mytoken1 = '1111')
        assert expected == actual, 'the text was not modified as expected'

    def test_expand_multiple_occurrences(self):
        """Test expand() replaces multiple tokens in text."""
        text = 'The following token ${mytoken1} should be replaced, as should ${mytoken2}.'
        expected = 'The following token 1111 should be replaced, as should 2222.'
        actual = Utility.expand(text, mytoken1 = '1111', mytoken2 = '2222')
        assert expected == actual, 'the text was not modified as expected'

    def test_expand_repeated_occurrences(self):
        """Test expand() replaces repeated tokens in text."""
        text = 'The following token ${mytoken1} should be replaced, as should ${mytoken1}.'
        expected = 'The following token 1111 should be replaced, as should 1111.'
        actual = Utility.expand(text, mytoken1 = '1111')
        assert expected == actual, 'the text was not modified as expected'



##############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ATLASCastorDataset.py,v 1.1 2008-07-17 16:41:18 moscicki Exp $
###############################################################################
# A simple ATLAS dataset
#
# ATLAS/ARDA

import os, re

from Ganga.GPIDev.Lib.Dataset import Dataset
from Ganga.GPIDev.Schema import *

from Ganga.Utility.logging import getLogger

from Ganga.Utility.Shell import Shell

logger = getLogger()
shell = Shell()
   
class ATLASCastorDataset(Dataset):
    '''ATLAS data as list of files on CASTOR'''
    
    _schema = Schema(Version(1,0), {
        'location' : SimpleItem(defvalue = '',
            doc = 'A directory on castor that contains the datasets in directories'),
        'pattern'  : SimpleItem(defvalue = '',
            doc = 'A regexp filter to select the correct files'),
        'dataset'  : SimpleItem(defvalue = '',
            doc = 'The name of the selected dataset'),
        'names'    : SimpleItem(defvalue = [], typelist=['str'], sequence = 1,
            doc = 'The selected file names')
    })
   
    _category = 'datasets'
    _name = 'ATLASCastorDataset'
   
    _exportmethods = ['set_dataset_type','list_datasets','get_dataset', 'get_dataset_filenames' ]

    _GUIPrefs = [ { 'attribute' : 'location',       'widget' : 'String' },
                  { 'attribute' : 'pattern',        'widget' : 'String' },
                  { 'attribute' : 'dataset',        'widget' : 'String' },
                  { 'attribute' : 'names',          'widget' : 'String_List' } ]
   
    def __init__(self):

        super(ATLASCastorDataset,self).__init__()
        self.set_dataset_type('mc11-aod')

    def set_dataset_type(self,name):
        '''Set the dataset type (rome-aod,rome-aod-merged, mc11-aod,csc11-aod)'''

        if name == 'rome-aod':
            self.location = '/castor/cern.ch/grid/atlas/datafiles/rome/recov10'
            self.pattern = '\S+\.\d+\.reco\S+\.\S+\.AOD\.pool\.root\S*'
        elif name == 'rome-aod-merged':
            self.location = '/castor/cern.ch/grid/atlas/datafiles/rome/recov10/merge'
            self.pattern = 'rome\.\d+\.merge\.\S+\.AOD\.pool\.root'
        elif name == 'csc11-aod':
            self.location = '/castor/cern.ch/atlas/csc/valiprod/sampleA/csc11'
            self.pattern = ''
        elif name == 'mc11-aod':
            self.location = '/castor/cern.ch/atlas/csc/valiprod/sampleA/mc11'
            self.pattern = ''
        else:
            logger.warning('Unknown dataset type %d. It should be rome-aod, rome-aod-merged, mc11-aod or csc11-aod.')
            self.location=''
            self.pattern=''

        self._setDirty()
   
    def list_datasets(self):
        '''List the datasets of a given type'''
     
        if not self.location:
            logger.error('No dataset location has been defined.')
            return

        datasets = []
        if self.location.find('valiprod') > 0:
            rc, output, m = shell.cmd1('nsls %s' % self.location,[0,1])
            for ds1 in re.findall('(\S+)\n',output):
                rc, output, m = shell.cmd1('nsls %s/%s/recon' % (self.location,ds1),[0,1])
                if rc == 1: continue
                datasets += [ '%s.recon.%s' % (ds1,ds2) for ds2 in re.findall('(\S+)\n',output) ]
 
        else:        
            rc, output, m = shell.cmd1('nsls %s' % self.location)
            datasets += [ ds for ds in re.findall('(\S+)\n',output) if ds.count('.') == 3 ]

        return datasets
         
    def get_dataset(self,dataset,start=0,end=-1):
       '''Get the actual files of a dataset.'''
      
       if not self.location:
           logger.error('No dataset location has been defined.')
           return
       
       if self.location.find('valiprod') > 0:
           ds = dataset.split('.')
           if len(ds) == 5:
              path = '%s/%s.%s/recon/%s' % (self.location,ds[1],ds[2],ds[4])
           elif len(ds) == 6:           
              path = '%s/%s.%s/recon/%s' % (self.location,ds[1],ds[2],ds[5])
           else:
              logger.error('Invalid dataset name %s',dataset)
              return
              
       else:
           path = '%s/%s' % (self.location,dataset)

       logger.info('Reading %s ...',path)

       rc, output, m = shell.cmd1('nsls %s' % path,[0,1])
       if rc == 1: 
           logger.error('Dataset %s does not exist at %s',dataset,self.location)
           return

       self.dataset = dataset

       if self.pattern:
           names = re.findall('(%s)\n' % self.pattern,output)
       else:
           names = re.findall('(\S+)\n',output)

       self.names = names[start:end]

       self._setDirty()
           
    def filenames(self):
        '''Get the full filename'''
      
        if self.location.find('valiprod') > 0:

            ds = self.dataset.split('.')
            if len(ds) == 5:
                path = '%s/%s.%s/recon/%s' % (self.location,ds[1],ds[2],ds[4])
            elif len(ds) == 6:
                path = '%s/%s.%s/recon/%s' % (self.location,ds[1],ds[2],ds[5])
            else:
              logger.error('Invalid dataset name %s',dataset)
              return []
                                                                                                                                 
            return [ 'rfio:%s/%s' % (path,name) for name in self.names ]
        else:
            return [ 'rfio:%s/%s/%s' % (self.location,self.dataset,name) for name in self.names ]

    @staticmethod
    def get_filenames(app):
        '''Retrieve the file names starting from an application object'''
      
        job=app._getRoot()
        if not job:
            logger.warning('Application object is not associated to a job.')
	    return []
	 
#       jobs without inputdata are allowed
	 
        if not job.inputdata: return []
      
        if not job.inputdata._name == 'ATLASCastorDataset':
            logger.warning('Dataset is not of type ATLASCastorDataset.')
	    return []

        return job.inputdata.filenames()

    def get_dataset_filenames(self):
        '''Retrieve the file names'''
      
        return self.filenames()



"""Core extractor IAction implementation.

The CoreExtractor class provides a generic extractor implementation.
 
"""

from GangaRobot.Framework import Utility
from GangaRobot.Lib.Base.BaseExtractor import BaseExtractor
from Ganga.GPI import *


class CoreExtractor(BaseExtractor):

    """Core extractor IAction implementation.
    
    An extractor which extracts generic data common to most Ganga jobs
    irrespective of application or backend used. See handlerunnode() and
    handlejobnode() for details.
    
    The extracted run node has the structure:
    <run>
        <core/>
        <job>
            <core/>
        </job>
        <job>
            <core/>
        </job>
        ...
    </run>
    
    This extractor can be reused by adding it to the chain of extractors
    initialised in the constructor of any implementation of BaseExtractor.
    e.g.
    def __init__(self):
        self.chain = [CoreExtractor(), self]
    
    """

    def handlerunnode(self, runnode, runid):
        """Extracts generic data for the given run.
        
        Keyword arguments:
        runnode -- An Extract.Node for the run.
        runid -- A UTC ID string which identifies the run.
        
        Example of nodes added to the run node:
        <core>
            <id>2007-06-22_13.17.51</id>
            <start-time>2007/06/22 13:17:51</start-time>
            <extract-time>2007/06/22 13:18:55</extract-time>
        </core>
        
        All nodes are guaranteed to be present. All times are in UTC.
        
        """
        corenode = runnode.addnode('core')
        corenode.addnode('id', runid)
        corenode.addnode('start-time', Utility.utctime(runid))
        corenode.addnode('extract-time', Utility.utctime())

    def handlejobnode(self, jobnode, job):
        """Extracts generic data for the given job.
        
        Keyword arguments:
        jobnode -- An Extract.Node for the given job.
        job -- A Ganga Job associated with the run.
        
        Example of nodes added to the job node:
        <core>
            <id>4</id>
            <name>Core_2007-06-22_13.17.51_0</name>
            <status>completed</status>
            <application>Executable</application>
            <backend>Local</backend>
            <backend-id>19253</backend-id>
            <backend-actualCE>lx09.hep.ph.ic.ac.uk</backend-actualCE>
        </core>
        
        All nodes are guaranteed to be present although their text values may be
        empty if querying the job returns no corresponding value.
        
        """
        corenode = jobnode.addnode('core')
        corenode.addnode('id', job.id)
        corenode.addnode('name', job.name)
        corenode.addnode('status', job.status)
        #FIXME: Add job timestamps when available
        corenode.addnode('application', typename(job.application))
        corenode.addnode('backend', typename(job.backend))
        backendid = None
        statusInfo = None
        if hasattr(job.backend,'statusInfo'):
            statusInfo = job.backend.statusInfo
        corenode.addnode('minor-status',statusInfo)
        if hasattr(job.backend, 'id'):
            backendid = job.backend.id 
        corenode.addnode('backend-id', backendid)
        backendactualCE = None
        if hasattr(job.backend, 'actualCE'):
            backendactualCE = job.backend.actualCE
        corenode.addnode('backend-actualCE', backendactualCE)



#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
import os
from GangaGaudi.Lib.RTHandlers.RunTimeHandlerUtils import master_sandbox_prepare, sandbox_prepare, script_generator
from GangaDirac.Lib.RTHandlers.DiracRTHUtils import dirac_inputdata, dirac_ouputdata, mangle_job_name, diracAPI_script_template, diracAPI_script_settings
from GangaGaudi.Lib.RTHandlers.GaudiRunTimeHandler import GaudiRunTimeHandler
from Ganga.GPIDev.Adapters.StandardJobConfig import StandardJobConfig
from Ganga.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatterns, getWNCodeForOutputPostprocessing
from Ganga.Utility.Config import getConfig
from Ganga.Utility.util import unique
from GangaDirac.Lib.Files.DiracFile import DiracFile
#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#


class GaudiDiracRunTimeHandler(GaudiRunTimeHandler):

    """The runtime handler to run Gaudi jobs on the Dirac backend"""

    def prepare(self, app, appsubconfig, appmasterconfig, jobmasterconfig):
        inputsandbox, outputsandbox = sandbox_prepare(
            app, appsubconfig, appmasterconfig, jobmasterconfig)
        input_data,   parametricinput_data = dirac_inputdata(app)

        job = app.getJobObject()
        outputfiles = [this_file.namePattern for this_file in job.outputfiles if isinstance(this_file, DiracFile)]

        gaudi_script_path = os.path.join(
            job.getInputWorkspace().getPath(), "gaudi-script.py")
        script_generator(gaudi_script_template(),
                         #remove_unreplaced = False,
                         outputfile_path=gaudi_script_path,
                         PLATFORM=app.platform,
                         COMMAND='gaudirun.py'  # ,
                         #OUTPUTFILESINJECTEDCODE = getWNCodeForOutputPostprocessing(job, '    ')
                         )

        dirac_script = script_generator(diracAPI_script_template(),
                                        DIRAC_IMPORT='from DIRAC.Interfaces.API.Dirac import Dirac',
                                        DIRAC_JOB_IMPORT='from DIRAC.Interfaces.API.Job import Job',
                                        DIRAC_OBJECT='Dirac()',
                                        JOB_OBJECT='Job()',
                                        NAME=mangle_job_name(app),
                                        EXE=gaudi_script_path,
                                        EXE_ARG_STR=' '.join(
                                            [str(arg) for arg in app.args]),
                                        EXE_LOG_FILE='Ganga_%s_%s.log' % (
                                            app.appname, app.version),
                                        INPUTDATA=input_data,
                                        PARAMETRIC_INPUTDATA=parametricinput_data,
                                        OUTPUT_SANDBOX=outputsandbox,
                                        OUTPUTDATA=list(outputfiles),
                                        OUTPUT_PATH="",  # job.fqid,
                                        OUTPUT_SE=getConfig(
                                            'DIRAC')['DiracOutputDataSE'],
                                        SETTINGS=diracAPI_script_settings(app),
                                        DIRAC_OPTS=job.backend.diracOpts,
                                        PLATFORM=app.platform,
                                        # leave the sandbox for altering later as needs
                                        # to be done in backend.submit to combine master.
                                        # Note only using 2 #s as auto-remove 3
                                        INPUT_SANDBOX='##INPUT_SANDBOX##'
                                        )

        return StandardJobConfig(dirac_script,
                                 inputbox=unique(inputsandbox),
                                 outputbox=unique(outputsandbox))


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
def gaudi_script_template(self):
    '''Creates the script that will be executed by DIRAC job. '''

    script_template = """#!/usr/bin/env python
'''Script to run Gaudi application'''

from os import curdir, system, environ, pathsep, sep, getcwd
from os.path import join
import sys

def prependEnv(key, value):
    if key in environ: value += (pathsep + environ[key])
    environ[key] = value

# Main
if __name__ == '__main__':

    prependEnv('LD_LIBRARY_PATH', getcwd() + '/lib')
    prependEnv('PYTHONPATH', getcwd() + '/InstallArea/python')
    prependEnv('PYTHONPATH', getcwd() + '/InstallArea/###PLATFORM###/python')

    rc = system('''###COMMAND### ''' + ' '.join(sys.argv))/256

    ###OUTPUTFILESINJECTEDCODE###
    sys.exit(rc)
"""
    return script_template


#\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\#
